from .. import register_map
from ..resources import ExternalResources, KeyTable, ResourceTable, ObjectTable, ObjectKeyTable, EntityTable
from ...build import ObjectMapper


@register_map(ExternalResources)
class ExternalResourcesMap(ObjectMapper):

    def construct_helper(self, name, parent_builder, table_cls, manager):
        """Create a new instance of table_cls with data from parent_builder[name].

           The DatasetBuilder for name is associated with data_type Data and container class Data,
           but users should use the more specific table_cls for these datasets.
        """
        parent = manager._get_proxy_builder(parent_builder)
        builder = parent_builder[name]
        src = builder.source
        oid = builder.attributes.get(self.spec.id_key())
        kwargs = dict(name=builder.name, data=builder.data)
        return self.__new_container__(table_cls, src, parent, oid, **kwargs)

    @ObjectMapper.constructor_arg('keys')
    def keys(self, builder, manager):
        return self.construct_helper('keys', builder, KeyTable, manager)

    @ObjectMapper.constructor_arg('resources')
    def resources(self, builder, manager):
        return self.construct_helper('resources', builder, ResourceTable, manager)

    @ObjectMapper.constructor_arg('entities')
    def entities(self, builder, manager):
        return self.construct_helper('entities', builder, EntityTable, manager)

    @ObjectMapper.constructor_arg('objects')
    def objects(self, builder, manager):
        return self.construct_helper('objects', builder, ObjectTable, manager)

    @ObjectMapper.constructor_arg('object_keys')
    def object_keys(self, builder, manager):
        return self.construct_helper('object_keys', builder, ObjectKeyTable, manager)



import types
from abc import abstractmethod
from collections import OrderedDict
from copy import deepcopy
from uuid import uuid4
from warnings import warn

import h5py
import numpy as np
import pandas as pd

from .data_utils import DataIO, append_data, extend_data
from .utils import docval, get_docval, getargs, ExtenderMeta, get_data_shape, popargs, LabelledDict


def _set_exp(cls):
    """Set a class as being experimental"""
    cls._experimental = True


def _exp_warn_msg(cls):
    """Generate a warning message experimental features"""
    pfx = cls
    if isinstance(cls, type):
        pfx = cls.__name__
    msg = ('%s is experimental -- it may be removed in the future and '
           'is not guaranteed to maintain backward compatibility') % pfx
    return msg


class AbstractContainer(metaclass=ExtenderMeta):
    # The name of the class attribute that subclasses use to autogenerate properties
    # This parameterization is supplied in case users would like to configure
    # the class attribute name to something domain-specific

    _experimental = False

    _fieldsname = '__fields__'

    _data_type_attr = 'data_type'

    # Subclasses use this class attribute to add properties to autogenerate
    # Autogenerated properties will store values in self.__field_values
    __fields__ = tuple()

    # This field is automatically set by __gather_fields before initialization.
    # It holds all the values in __fields__ for this class and its parent classes.
    __fieldsconf = tuple()

    _pconf_allowed_keys = {'name', 'doc', 'settable'}

    # Override the _setter factor function, so directives that apply to
    # Container do not get used on Data
    @classmethod
    def _setter(cls, field):
        """
        Make a setter function for creating a :py:func:`property`
        """
        name = field['name']

        if not field.get('settable', True):
            return None

        def setter(self, val):
            if val is None:
                return
            if name in self.fields:
                msg = "can't set attribute '%s' -- already set" % name
                raise AttributeError(msg)
            self.fields[name] = val

        return setter

    @classmethod
    def _getter(cls, field):
        """
        Make a getter function for creating a :py:func:`property`
        """
        doc = field.get('doc')
        name = field['name']

        def getter(self):
            return self.fields.get(name)

        setattr(getter, '__doc__', doc)
        return getter

    @staticmethod
    def _check_field_spec(field):
        """
        A helper function for __gather_fields to make sure we are always working
        with a dict specification and that the specification contains the correct keys
        """
        tmp = field
        if isinstance(tmp, dict):
            if 'name' not in tmp:
                raise ValueError("must specify 'name' if using dict in __fields__")
        else:
            tmp = {'name': tmp}
        return tmp

    @classmethod
    def _check_field_spec_keys(cls, field_conf):
        for k in field_conf:
            if k not in cls._pconf_allowed_keys:
                msg = ("Unrecognized key '%s' in %s config '%s' on %s"
                       % (k, cls._fieldsname, field_conf['name'], cls.__name__))
                raise ValueError(msg)

    @classmethod
    def _get_fields(cls):
        return getattr(cls, cls._fieldsname)

    @classmethod
    def _set_fields(cls, value):
        return setattr(cls, cls._fieldsname, value)

    @classmethod
    def get_fields_conf(cls):
        return cls.__fieldsconf

    @ExtenderMeta.pre_init
    def __gather_fields(cls, name, bases, classdict):
        '''
        This classmethod will be called during class declaration in the metaclass to automatically
        create setters and getters for fields that need to be exported
        '''
        fields = cls._get_fields()
        if not isinstance(fields, tuple):
            msg = "'%s' must be of type tuple" % cls._fieldsname
            raise TypeError(msg)

        # check field specs and create map from field name to field conf dictionary
        fields_dict = OrderedDict()
        for f in fields:
            pconf = cls._check_field_spec(f)
            cls._check_field_spec_keys(pconf)
            fields_dict[pconf['name']] = pconf
        all_fields_conf = list(fields_dict.values())

        # check whether this class overrides __fields__
        if len(bases):
            # find highest base class that is an AbstractContainer (parent is higher than children)
            base_cls = None
            for base_cls in reversed(bases):
                if issubclass(base_cls, AbstractContainer):
                    break

            base_fields = base_cls._get_fields()  # tuple of field names from base class
            if base_fields is not fields:
                # check whether new fields spec already exists in base class
                fields_to_remove_from_base = list()
                for field_name in fields_dict:
                    if field_name in base_fields:
                        fields_to_remove_from_base.append(field_name)
                # prepend field specs from base class to fields list of this class
                # but only field specs that are not redefined in this class
                base_fields_conf = base_cls.get_fields_conf()  # tuple of fields configurations from base class
                base_fields_conf_to_add = list()
                for pconf in base_fields_conf:
                    if pconf['name'] not in fields_to_remove_from_base:
                        base_fields_conf_to_add.append(pconf)
                all_fields_conf[0:0] = base_fields_conf_to_add

        # create getter and setter if attribute does not already exist
        # if 'doc' not specified in __fields__, use doc from docval of __init__
        docs = {dv['name']: dv['doc'] for dv in get_docval(cls.__init__)}
        for field_conf in all_fields_conf:
            pname = field_conf['name']
            field_conf.setdefault('doc', docs.get(pname))
            if not hasattr(cls, pname):
                setattr(cls, pname, property(cls._getter(field_conf), cls._setter(field_conf)))

        cls._set_fields(tuple(field_conf['name'] for field_conf in all_fields_conf))
        cls.__fieldsconf = tuple(all_fields_conf)

    def __new__(cls, *args, **kwargs):
        """
        Static method of the object class called by Python to create the object first and then
         __init__() is called to initialize the object's attributes.

        NOTE: this method is called directly from ObjectMapper.__new_container__ during the process of
        constructing the object from builders that are read from a file.
        """
        inst = super().__new__(cls)
        if cls._experimental:
            warn(_exp_warn_msg(cls))
        inst.__container_source = kwargs.pop('container_source', None)
        inst.__parent = None
        inst.__children = list()
        inst.__modified = True
        inst.__object_id = kwargs.pop('object_id', str(uuid4()))
        # this variable is being passed in from ObjectMapper.__new_container__ and is
        # reset to False in that method after the object has been initialized by __init__
        inst._in_construct_mode = kwargs.pop('in_construct_mode', False)
        inst.parent = kwargs.pop('parent', None)
        return inst

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this container'})
    def __init__(self, **kwargs):
        name = getargs('name', kwargs)
        if '/' in name:
            raise ValueError("name '" + name + "' cannot contain '/'")
        self.__name = name
        self.__field_values = dict()

    @property
    def name(self):
        '''
        The name of this Container
        '''
        return self.__name

    @docval({'name': 'data_type', 'type': str, 'doc': 'the data_type to search for', 'default': None})
    def get_ancestor(self, **kwargs):
        """
        Traverse parent hierarchy and return first instance of the specified data_type
        """
        data_type = getargs('data_type', kwargs)
        if data_type is None:
            return self.parent
        p = self.parent
        while p is not None:
            if getattr(p, p._data_type_attr) == data_type:
                return p
            p = p.parent
        return None

    @property
    def fields(self):
        return self.__field_values

    @property
    def object_id(self):
        if self.__object_id is None:
            self.__object_id = str(uuid4())
        return self.__object_id

    @docval({'name': 'recurse', 'type': bool,
             'doc': "whether or not to change the object ID of this container's children", 'default': True})
    def generate_new_id(self, **kwargs):
        """Changes the object ID of this Container and all of its children to a new UUID string."""
        recurse = getargs('recurse', kwargs)
        self.__object_id = str(uuid4())
        self.set_modified()
        if recurse:
            for c in self.children:
                c.generate_new_id(**kwargs)

    @property
    def modified(self):
        return self.__modified

    @docval({'name': 'modified', 'type': bool,
             'doc': 'whether or not this Container has been modified', 'default': True})
    def set_modified(self, **kwargs):
        modified = getargs('modified', kwargs)
        self.__modified = modified
        if modified and isinstance(self.parent, Container):
            self.parent.set_modified()

    @property
    def children(self):
        return tuple(self.__children)

    @docval({'name': 'child', 'type': 'Container',
             'doc': 'the child Container for this Container', 'default': None})
    def add_child(self, **kwargs):
        warn(DeprecationWarning('add_child is deprecated. Set the parent attribute instead.'))
        child = getargs('child', kwargs)
        if child is not None:
            # if child.parent is a Container, then the mismatch between child.parent and parent
            # is used to make a soft/external link from the parent to a child elsewhere
            # if child.parent is not a Container, it is either None or a Proxy and should be set to self
            if not isinstance(child.parent, AbstractContainer):
                # actually add the child to the parent in parent setter
                child.parent = self
        else:
            warn('Cannot add None as child to a container %s' % self.name)

    @classmethod
    def type_hierarchy(cls):
        return cls.__mro__

    @property
    def container_source(self):
        '''
        The source of this Container
        '''
        return self.__container_source

    @container_source.setter
    def container_source(self, source):
        if self.__container_source is not None:
            raise Exception('cannot reassign container_source')
        self.__container_source = source

    @property
    def parent(self):
        '''
        The parent Container of this Container
        '''
        # do it this way because __parent may not exist yet (not set in constructor)
        return getattr(self, '_AbstractContainer__parent', None)

    @parent.setter
    def parent(self, parent_container):
        if self.parent is parent_container:
            return

        if self.parent is not None:
            if isinstance(self.parent, AbstractContainer):
                raise ValueError(('Cannot reassign parent to Container: %s. '
                                  'Parent is already: %s.' % (repr(self), repr(self.parent))))
            else:
                if parent_container is None:
                    raise ValueError("Got None for parent of '%s' - cannot overwrite Proxy with NoneType" % repr(self))
                # NOTE this assumes isinstance(parent_container, Proxy) but we get a circular import
                # if we try to do that
                if self.parent.matches(parent_container):
                    self.__parent = parent_container
                    parent_container.__children.append(self)
                    parent_container.set_modified()
                else:
                    self.__parent.add_candidate(parent_container)
        else:
            self.__parent = parent_container
            if isinstance(parent_container, Container):
                parent_container.__children.append(self)
                parent_container.set_modified()

    def _remove_child(self, child):
        """Remove a child Container. Intended for use in subclasses that allow dynamic addition of child Containers."""
        if not isinstance(child, AbstractContainer):
            raise ValueError('Cannot remove non-AbstractContainer object from children.')
        if child not in self.children:
            raise ValueError("%s '%s' is not a child of %s '%s'." % (child.__class__.__name__, child.name,
                                                                     self.__class__.__name__, self.name))
        child.__parent = None
        self.__children.remove(child)
        child.set_modified()
        self.set_modified()

    def reset_parent(self):
        """Reset the parent of this Container to None and remove the Container from the children of its parent.

        Use with caution. This can result in orphaned containers and broken links.
        """
        if self.parent is None:
            return
        elif isinstance(self.parent, AbstractContainer):
            self.parent._remove_child(self)
        else:
            raise ValueError("Cannot reset parent when parent is not an AbstractContainer: %s" % repr(self.parent))


class Container(AbstractContainer):
    """A container that can contain other containers and has special functionality for printing."""

    _pconf_allowed_keys = {'name', 'child', 'required_name', 'doc', 'settable'}

    @classmethod
    def _setter(cls, field):
        """Returns a list of setter functions for the given field to be added to the class during class declaration."""
        super_setter = AbstractContainer._setter(field)
        ret = [super_setter]
        # create setter with check for required name
        # the AbstractContainer that is passed to the setter must have name = required_name
        if field.get('required_name', None) is not None:
            required_name = field['required_name']
            idx1 = len(ret) - 1

            def container_setter(self, val):
                if val is not None:
                    if not isinstance(val, AbstractContainer):
                        msg = ("Field '%s' on %s has a required name and must be a subclass of AbstractContainer."
                               % (field['name'], self.__class__.__name__))
                        raise ValueError(msg)
                    if val.name != required_name:
                        msg = ("Field '%s' on %s must be named '%s'."
                               % (field['name'], self.__class__.__name__, required_name))
                        raise ValueError(msg)
                ret[idx1](self, val)  # call the previous setter

            ret.append(container_setter)

        # create setter that accepts a value or tuple, list, or dict or values and sets the value's parent to self
        if field.get('child', False):
            idx2 = len(ret) - 1

            def container_setter(self, val):
                ret[idx2](self, val)  # call the previous setter
                if val is not None:
                    if isinstance(val, (tuple, list)):
                        pass
                    elif isinstance(val, dict):
                        val = val.values()
                    else:
                        val = [val]
                    for v in val:
                        if not isinstance(v.parent, Container):
                            v.parent = self
                        else:
                            # the ObjectMapper will create a link from self (parent) to v (child with existing parent)
                            # still need to mark self as modified
                            self.set_modified()

            ret.append(container_setter)
        return ret[-1]  # return the last setter (which should call the previous setters, if applicable)

    def __repr__(self):
        cls = self.__class__
        template = "%s %s.%s at 0x%d" % (self.name, cls.__module__, cls.__name__, id(self))
        if len(self.fields):
            template += "\nFields:\n"
        for k in sorted(self.fields):  # sorted to enable tests
            v = self.fields[k]
            # if isinstance(v, DataIO) or not hasattr(v, '__len__') or len(v) > 0:
            if hasattr(v, '__len__'):
                if isinstance(v, (np.ndarray, list, tuple)):
                    if len(v) > 0:
                        template += "  {}: {}\n".format(k, self.__smart_str(v, 1))
                elif v:
                    template += "  {}: {}\n".format(k, self.__smart_str(v, 1))
            else:
                template += "  {}: {}\n".format(k, v)
        return template

    @staticmethod
    def __smart_str(v, num_indent):
        """
        Print compact string representation of data.

        If v is a list, try to print it using numpy. This will condense the string
        representation of datasets with many elements. If that doesn't work, just print the list.

        If v is a dictionary, print the name and type of each element

        If v is a set, print it sorted

        If v is a neurodata_type, print the name of type

        Otherwise, use the built-in str()
        Parameters
        ----------
        v

        Returns
        -------
        str

        """

        if isinstance(v, list) or isinstance(v, tuple):
            if len(v) and isinstance(v[0], AbstractContainer):
                return Container.__smart_str_list(v, num_indent, '(')
            try:
                return str(np.asarray(v))
            except ValueError:
                return Container.__smart_str_list(v, num_indent, '(')
        elif isinstance(v, dict):
            return Container.__smart_str_dict(v, num_indent)
        elif isinstance(v, set):
            return Container.__smart_str_list(sorted(list(v)), num_indent, '{')
        elif isinstance(v, AbstractContainer):
            return "{} {}".format(getattr(v, 'name'), type(v))
        else:
            return str(v)

    @staticmethod
    def __smart_str_list(str_list, num_indent, left_br):
        if left_br == '(':
            right_br = ')'
        if left_br == '{':
            right_br = '}'
        if len(str_list) == 0:
            return left_br + ' ' + right_br
        indent = num_indent * 2 * ' '
        indent_in = (num_indent + 1) * 2 * ' '
        out = left_br
        for v in str_list[:-1]:
            out += '\n' + indent_in + Container.__smart_str(v, num_indent + 1) + ','
        if str_list:
            out += '\n' + indent_in + Container.__smart_str(str_list[-1], num_indent + 1)
        out += '\n' + indent + right_br
        return out

    @staticmethod
    def __smart_str_dict(d, num_indent):
        left_br = '{'
        right_br = '}'
        if len(d) == 0:
            return left_br + ' ' + right_br
        indent = num_indent * 2 * ' '
        indent_in = (num_indent + 1) * 2 * ' '
        out = left_br
        keys = sorted(list(d.keys()))
        for k in keys[:-1]:
            out += '\n' + indent_in + Container.__smart_str(k, num_indent + 1) + ' ' + str(type(d[k])) + ','
        if keys:
            out += '\n' + indent_in + Container.__smart_str(keys[-1], num_indent + 1) + ' ' + str(type(d[keys[-1]]))
        out += '\n' + indent + right_br
        return out


class Data(AbstractContainer):
    """
    A class for representing dataset containers
    """

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this container'},
            {'name': 'data', 'type': ('scalar_data', 'array_data', 'data'), 'doc': 'the source of the data'})
    def __init__(self, **kwargs):
        data = popargs('data', kwargs)
        super().__init__(**kwargs)
        self.__data = data

    @property
    def data(self):
        return self.__data

    @property
    def shape(self):
        """
        Get the shape of the data represented by this container
        :return: Shape tuple
        :rtype: tuple of ints
        """
        return get_data_shape(self.__data)

    @docval({'name': 'dataio', 'type': DataIO, 'doc': 'the DataIO to apply to the data held by this Data'})
    def set_dataio(self, **kwargs):
        """
        Apply DataIO object to the data held by this Data object
        """
        dataio = getargs('dataio', kwargs)
        dataio.data = self.__data
        self.__data = dataio

    @docval({'name': 'func', 'type': types.FunctionType, 'doc': 'a function to transform *data*'})
    def transform(self, **kwargs):
        """
        Transform data from the current underlying state.

        This function can be used to permanently load data from disk, or convert to a different
        representation, such as a torch.Tensor
        """
        func = getargs('func', kwargs)
        self.__data = func(self.__data)
        return self

    def __bool__(self):
        if self.data is not None:
            if isinstance(self.data, (np.ndarray, tuple, list)):
                return len(self.data) != 0
            if self.data:
                return True
        return False

    def __len__(self):
        return len(self.__data)

    def __getitem__(self, args):
        return self.get(args)

    def get(self, args):
        if isinstance(self.data, (tuple, list)) and isinstance(args, (tuple, list, np.ndarray)):
            return [self.data[i] for i in args]
        if isinstance(self.data, h5py.Dataset) and isinstance(args, np.ndarray):
            # This is needed for h5py 2.9 compatability
            args = args.tolist()
        return self.data[args]

    def append(self, arg):
        self.__data = append_data(self.__data, arg)

    def extend(self, arg):
        """
        The extend_data method adds all the elements of the iterable arg to the
        end of the data of this Data container.

        :param arg: The iterable to add to the end of this VectorData
        """
        self.__data = extend_data(self.__data, arg)


class DataRegion(Data):

    @property
    @abstractmethod
    def data(self):
        '''
        The target data that this region applies to
        '''
        pass

    @property
    @abstractmethod
    def region(self):
        '''
        The region that indexes into data e.g. slice or list of indices
        '''
        pass


class MultiContainerInterface(Container):
    """Class that dynamically defines methods to support a Container holding multiple Containers of the same type.

    To use, extend this class and create a dictionary as a class attribute with any of the following keys:
    * 'attr' to name the attribute that stores the Container instances
    * 'type' to provide the Container object type (type or list/tuple of types, type can be a docval macro)
    * 'add' to name the method for adding Container instances
    * 'get' to name the method for getting Container instances
    * 'create' to name the method for creating Container instances (only if a single type is specified)

    If the attribute does not exist in the class, it will be generated. If it does exist, it should behave like a dict.

    The keys 'attr', 'type', and 'add' are required.
    """

    def __new__(cls, *args, **kwargs):
        if cls is MultiContainerInterface:
            raise TypeError("Can't instantiate class MultiContainerInterface.")
        if not hasattr(cls, '__clsconf__'):
            raise TypeError("MultiContainerInterface subclass %s is missing __clsconf__ attribute. Please check that "
                            "the class is properly defined." % cls.__name__)
        return super().__new__(cls, *args, **kwargs)

    @staticmethod
    def __add_article(noun):
        if isinstance(noun, tuple):
            noun = noun[0]
        if isinstance(noun, type):
            noun = noun.__name__
        if noun[0] in ('aeiouAEIOU'):
            return 'an %s' % noun
        return 'a %s' % noun

    @staticmethod
    def __join(argtype):
        """Return a grammatical string representation of a list or tuple of classes or text.

        Examples:
        cls.__join(Container) returns "Container"
        cls.__join((Container, )) returns "Container"
        cls.__join((Container, Data)) returns "Container or Data"
        cls.__join((Container, Data, Subcontainer)) returns "Container, Data, or Subcontainer"
        """

        def tostr(x):
            return x.__name__ if isinstance(x, type) else x

        if isinstance(argtype, (list, tuple)):
            args_str = [tostr(x) for x in argtype]
            if len(args_str) == 1:
                return args_str[0]
            if len(args_str) == 2:
                return " or ".join(tostr(x) for x in args_str)
            else:
                return ", ".join(tostr(x) for x in args_str[:-1]) + ', or ' + args_str[-1]
        else:
            return tostr(argtype)

    @classmethod
    def __make_get(cls, func_name, attr_name, container_type):
        doc = "Get %s from this %s" % (cls.__add_article(container_type), cls.__name__)

        @docval({'name': 'name', 'type': str, 'doc': 'the name of the %s' % cls.__join(container_type),
                 'default': None},
                rtype=container_type, returns='the %s with the given name' % cls.__join(container_type),
                func_name=func_name, doc=doc)
        def _func(self, **kwargs):
            name = getargs('name', kwargs)
            d = getattr(self, attr_name)
            ret = None
            if name is None:
                if len(d) > 1:
                    msg = ("More than one element in %s of %s '%s' -- must specify a name."
                           % (attr_name, cls.__name__, self.name))
                    raise ValueError(msg)
                elif len(d) == 0:
                    msg = "%s of %s '%s' is empty." % (attr_name, cls.__name__, self.name)
                    raise ValueError(msg)
                else:  # only one item in dict
                    for v in d.values():
                        ret = v
            else:
                ret = d.get(name)
                if ret is None:
                    msg = "'%s' not found in %s of %s '%s'." % (name, attr_name, cls.__name__, self.name)
                    raise KeyError(msg)
            return ret

        return _func

    @classmethod
    def __make_getitem(cls, attr_name, container_type):
        doc = "Get %s from this %s" % (cls.__add_article(container_type), cls.__name__)

        @docval({'name': 'name', 'type': str, 'doc': 'the name of the %s' % cls.__join(container_type),
                 'default': None},
                rtype=container_type, returns='the %s with the given name' % cls.__join(container_type),
                func_name='__getitem__', doc=doc)
        def _func(self, **kwargs):
            # NOTE this is the same code as the getter but with different error messages
            name = getargs('name', kwargs)
            d = getattr(self, attr_name)
            ret = None
            if name is None:
                if len(d) > 1:
                    msg = ("More than one %s in %s '%s' -- must specify a name."
                           % (cls.__join(container_type), cls.__name__, self.name))
                    raise ValueError(msg)
                elif len(d) == 0:
                    msg = "%s '%s' is empty." % (cls.__name__, self.name)
                    raise ValueError(msg)
                else:  # only one item in dict
                    for v in d.values():
                        ret = v
            else:
                ret = d.get(name)
                if ret is None:
                    msg = "'%s' not found in %s '%s'." % (name, cls.__name__, self.name)
                    raise KeyError(msg)
            return ret

        return _func

    @classmethod
    def __make_add(cls, func_name, attr_name, container_type):
        doc = "Add one or multiple %s objects to this %s" % (cls.__join(container_type), cls.__name__)

        @docval({'name': attr_name, 'type': (list, tuple, dict, container_type),
                 'doc': 'one or multiple %s objects to add to this %s' % (cls.__join(container_type),  cls.__name__)},
                func_name=func_name, doc=doc)
        def _func(self, **kwargs):
            container = getargs(attr_name, kwargs)
            if isinstance(container, container_type):
                containers = [container]
            elif isinstance(container, dict):
                containers = container.values()
            else:
                containers = container
            d = getattr(self, attr_name)
            for tmp in containers:
                if not isinstance(tmp.parent, Container):
                    tmp.parent = self
                else:
                    # the ObjectMapper will create a link from self (parent) to tmp (child with existing parent)
                    # still need to mark self as modified
                    self.set_modified()
                if tmp.name in d:
                    msg = "'%s' already exists in %s '%s'" % (tmp.name, cls.__name__, self.name)
                    raise ValueError(msg)
                d[tmp.name] = tmp
            return container

        return _func

    @classmethod
    def __make_create(cls, func_name, add_name, container_type):
        doc = "Create %s object and add it to this %s" % (cls.__add_article(container_type), cls.__name__)

        @docval(*get_docval(container_type.__init__), func_name=func_name, doc=doc,
                returns="the %s object that was created" % cls.__join(container_type), rtype=container_type)
        def _func(self, **kwargs):
            ret = container_type(**kwargs)
            getattr(self, add_name)(ret)
            return ret

        return _func

    @classmethod
    def __make_constructor(cls, clsconf):
        args = list()
        for conf in clsconf:
            attr_name = conf['attr']
            container_type = conf['type']
            args.append({'name': attr_name, 'type': (list, tuple, dict, container_type),
                         'doc': '%s to store in this interface' % cls.__join(container_type), 'default': dict()})

        args.append({'name': 'name', 'type': str, 'doc': 'the name of this container', 'default': cls.__name__})

        @docval(*args, func_name='__init__')
        def _func(self, **kwargs):
            super().__init__(name=kwargs['name'])
            for conf in clsconf:
                attr_name = conf['attr']
                add_name = conf['add']
                container = popargs(attr_name, kwargs)
                add = getattr(self, add_name)
                add(container)

        return _func

    @classmethod
    def __make_getter(cls, attr):
        """Make a getter function for creating a :py:func:`property`"""

        def _func(self):
            # initialize the field to an empty labeled dict if it has not yet been
            # do this here to avoid creating default __init__ which may or may not be overridden in
            # custom classes and dynamically generated classes
            if attr not in self.fields:
                def _remove_child(child):
                    if child.parent is self:
                        self._remove_child(child)
                self.fields[attr] = LabelledDict(attr, remove_callable=_remove_child)

            return self.fields.get(attr)

        return _func

    @classmethod
    def __make_setter(cls, add_name):
        """Make a setter function for creating a :py:func:`property`"""

        @docval({'name': 'val', 'type': (list, tuple, dict), 'doc': 'the sub items to add', 'default': None})
        def _func(self, **kwargs):
            val = getargs('val', kwargs)
            if val is None:
                return
            getattr(self, add_name)(val)

        return _func

    @ExtenderMeta.pre_init
    def __build_class(cls, name, bases, classdict):
        """Verify __clsconf__ and create methods based on __clsconf__.
        This method is called prior to __new__ and __init__ during class declaration in the metaclass.
        """
        if not hasattr(cls, '__clsconf__'):
            return

        multi = False
        if isinstance(cls.__clsconf__, dict):
            clsconf = [cls.__clsconf__]
        elif isinstance(cls.__clsconf__, list):
            multi = True
            clsconf = cls.__clsconf__
        else:
            raise TypeError("'__clsconf__' for MultiContainerInterface subclass %s must be a dict or a list of "
                            "dicts." % cls.__name__)

        for conf_index, conf_dict in enumerate(clsconf):
            cls.__build_conf_methods(conf_dict, conf_index, multi)

        # make __getitem__ (square bracket access) only if one conf type is defined
        if len(clsconf) == 1:
            attr = clsconf[0].get('attr')
            container_type = clsconf[0].get('type')
            setattr(cls, '__getitem__', cls.__make_getitem(attr, container_type))

        # create the constructor, only if it has not been overridden
        # i.e. it is the same method as the parent class constructor
        if '__init__' not in cls.__dict__:
            setattr(cls, '__init__', cls.__make_constructor(clsconf))

    @classmethod
    def __build_conf_methods(cls, conf_dict, conf_index, multi):
        # get add method name
        add = conf_dict.get('add')
        if add is None:
            msg = "MultiContainerInterface subclass %s is missing 'add' key in __clsconf__" % cls.__name__
            if multi:
                msg += " at index %d" % conf_index
            raise ValueError(msg)

        # get container attribute name
        attr = conf_dict.get('attr')
        if attr is None:
            msg = "MultiContainerInterface subclass %s is missing 'attr' key in __clsconf__" % cls.__name__
            if multi:
                msg += " at index %d" % conf_index
            raise ValueError(msg)

        # get container type
        container_type = conf_dict.get('type')
        if container_type is None:
            msg = "MultiContainerInterface subclass %s is missing 'type' key in __clsconf__" % cls.__name__
            if multi:
                msg += " at index %d" % conf_index
            raise ValueError(msg)

        # create property with the name given in 'attr' only if the attribute is not already defined
        if not hasattr(cls, attr):
            getter = cls.__make_getter(attr)
            setter = cls.__make_setter(add)
            doc = "a dictionary containing the %s in this %s" % (cls.__join(container_type), cls.__name__)
            setattr(cls, attr, property(getter, setter, None, doc))

        # create the add method
        setattr(cls, add, cls.__make_add(add, attr, container_type))

        # create the create method, only if a single container type is specified
        create = conf_dict.get('create')
        if create is not None:
            if isinstance(container_type, type):
                setattr(cls, create, cls.__make_create(create, add, container_type))
            else:
                msg = ("Cannot specify 'create' key in __clsconf__ for MultiContainerInterface subclass %s "
                       "when 'type' key is not a single type") % cls.__name__
                if multi:
                    msg += " at index %d" % conf_index
                raise ValueError(msg)

        # create the get method
        get = conf_dict.get('get')
        if get is not None:
            setattr(cls, get, cls.__make_get(get, attr, container_type))


class Row(object, metaclass=ExtenderMeta):
    """
    A class for representing rows from a Table.

    The Table class can be indicated with the __table__. Doing so
    will set constructor arguments for the Row class and ensure that
    Row.idx is set appropriately when a Row is added to the Table. It will
    also add functionality to the Table class for getting Row objects.

    Note, the Row class is not needed for working with Table objects. This
    is merely convenience functionality for working with Tables.
    """

    __table__ = None

    @property
    def idx(self):
        """The index of this row in its respective Table"""
        return self.__idx

    @idx.setter
    def idx(self, val):
        if self.__idx is None:
            self.__idx = val
        else:
            raise ValueError("cannot reset the ID of a row object")

    @property
    def table(self):
        """The Table this Row comes from"""
        return self.__table

    @table.setter
    def table(self, val):
        if val is not None:
            self.__table = val
        if self.idx is None:
            self.idx = self.__table.add_row(**self.todict())

    @ExtenderMeta.pre_init
    def __build_row_class(cls, name, bases, classdict):
        table_cls = getattr(cls, '__table__', None)
        if table_cls is not None:
            columns = getattr(table_cls, '__columns__')
            if cls.__init__ == bases[-1].__init__:  # check if __init__ is overridden
                columns = deepcopy(columns)
                func_args = list()
                for col in columns:
                    func_args.append(col)
                func_args.append({'name': 'table', 'type': Table, 'default': None,
                                  'help': 'the table this row is from'})
                func_args.append({'name': 'idx', 'type': int, 'default': None,
                                  'help': 'the index for this row'})

                @docval(*func_args)
                def __init__(self, **kwargs):
                    super(cls, self).__init__()
                    table, idx = popargs('table', 'idx', kwargs)
                    self.__keys = list()
                    self.__idx = None
                    self.__table = None
                    for k, v in kwargs.items():
                        self.__keys.append(k)
                        setattr(self, k, v)
                    self.idx = idx
                    self.table = table

                setattr(cls, '__init__', __init__)

                def todict(self):
                    return {k: getattr(self, k) for k in self.__keys}

                setattr(cls, 'todict', todict)

            # set this so Table.row gets set when a Table is instantiated
            table_cls.__rowclass__ = cls
        else:
            if bases != (object,):
                raise ValueError('__table__ must be set if sub-classing Row')

    def __eq__(self, other):
        return self.idx == other.idx and self.table is other.table

    def __str__(self):
        return "Row(%i, %s) = %s" % (self.idx, self.table.name, str(self.todict()))


class RowGetter:
    """
    A simple class for providing __getitem__ functionality that returns
    Row objects to a Table.
    """

    def __init__(self, table):
        self.table = table
        self.cache = dict()

    def __getitem__(self, idx):
        ret = self.cache.get(idx)
        if ret is None:
            row = self.table[idx]
            ret = self.table.__rowclass__(*row, table=self.table, idx=idx)
            self.cache[idx] = ret
        return ret


class Table(Data):
    r'''
    Subclasses should specify the class attribute \_\_columns\_\_.

    This should be a list of dictionaries with the following keys:

    - ``name``            the column name
    - ``type``            the type of data in this column
    - ``doc``             a brief description of what gets stored in this column

    For reference, this list of dictionaries will be used with docval to autogenerate
    the ``add_row`` method for adding data to this table.

    If \_\_columns\_\_ is not specified, no custom ``add_row`` method will be added.

    The class attribute __defaultname__ can also be set to specify a default name
    for the table class. If \_\_defaultname\_\_ is not specified, then ``name`` will
    need to be specified when the class is instantiated.

    A Table class can be paired with a Row class for conveniently working with rows of
    a Table. This pairing must be indicated in the Row class implementation. See Row
    for more details.
    '''

    # This class attribute is used to indicate which Row class should be used when
    # adding RowGetter functionality to the Table.
    __rowclass__ = None

    @ExtenderMeta.pre_init
    def __build_table_class(cls, name, bases, classdict):
        if hasattr(cls, '__columns__'):
            columns = getattr(cls, '__columns__')

            idx = dict()
            for i, col in enumerate(columns):
                idx[col['name']] = i
            setattr(cls, '__colidx__', idx)

            if cls.__init__ == bases[-1].__init__:  # check if __init__ is overridden
                name = {'name': 'name', 'type': str, 'doc': 'the name of this table'}
                defname = getattr(cls, '__defaultname__', None)
                if defname is not None:
                    name['default'] = defname  # override the name with the default name if present

                @docval(name,
                        {'name': 'data', 'type': ('array_data', 'data'), 'doc': 'the data in this table',
                         'default': list()})
                def __init__(self, **kwargs):
                    name, data = getargs('name', 'data', kwargs)
                    colnames = [i['name'] for i in columns]
                    super(cls, self).__init__(colnames, name, data)

                setattr(cls, '__init__', __init__)

            if cls.add_row == bases[-1].add_row:  # check if add_row is overridden

                @docval(*columns)
                def add_row(self, **kwargs):
                    return super(cls, self).add_row(kwargs)

                setattr(cls, 'add_row', add_row)

    @docval({'name': 'columns', 'type': (list, tuple), 'doc': 'a list of the columns in this table'},
            {'name': 'name', 'type': str, 'doc': 'the name of this container'},
            {'name': 'data', 'type': ('array_data', 'data'), 'doc': 'the source of the data', 'default': list()})
    def __init__(self, **kwargs):
        self.__columns = tuple(popargs('columns', kwargs))
        self.__col_index = {name: idx for idx, name in enumerate(self.__columns)}
        if getattr(self, '__rowclass__') is not None:
            self.row = RowGetter(self)
        super().__init__(**kwargs)

    @property
    def columns(self):
        return self.__columns

    @docval({'name': 'values', 'type': dict, 'doc': 'the values for each column'})
    def add_row(self, **kwargs):
        values = getargs('values', kwargs)
        if not isinstance(self.data, list):
            msg = 'Cannot append row to %s' % type(self.data)
            raise ValueError(msg)
        ret = len(self.data)
        row = [values[col] for col in self.columns]
        row = [v.idx if isinstance(v, Row) else v for v in row]
        self.data.append(tuple(row))
        return ret

    def which(self, **kwargs):
        '''
        Query a table
        '''
        if len(kwargs) != 1:
            raise ValueError("only one column can be queried")
        colname, value = kwargs.popitem()
        idx = self.__colidx__.get(colname)
        if idx is None:
            msg = "no '%s' column in %s" % (colname, self.__class__.__name__)
            raise KeyError(msg)
        ret = list()
        for i in range(len(self.data)):
            row = self.data[i]
            row_val = row[idx]
            if row_val == value:
                ret.append(i)
        return ret

    def __len__(self):
        return len(self.data)

    def __getitem__(self, args):
        idx = args
        col = None
        if isinstance(args, tuple):
            idx = args[1]
            if isinstance(args[0], str):
                col = self.__col_index.get(args[0])
            elif isinstance(args[0], int):
                col = args[0]
            else:
                raise KeyError('first argument must be a column name or index')
            return self.data[idx][col]
        elif isinstance(args, str):
            col = self.__col_index.get(args)
            if col is None:
                raise KeyError(args)
            return [row[col] for row in self.data]
        else:
            return self.data[idx]

    def to_dataframe(self):
        '''Produce a pandas DataFrame containing this table's data.
        '''

        data = {colname: self[colname] for ii, colname in enumerate(self.columns)}
        return pd.DataFrame(data)

    @classmethod
    @docval(
        {'name': 'df', 'type': pd.DataFrame, 'doc': 'input data'},
        {'name': 'name', 'type': str, 'doc': 'the name of this container', 'default': None},
        {
            'name': 'extra_ok',
            'type': bool,
            'doc': 'accept (and ignore) unexpected columns on the input dataframe',
            'default': False
        },
    )
    def from_dataframe(cls, **kwargs):
        '''Construct an instance of Table (or a subclass) from a pandas DataFrame. The columns of the dataframe
        should match the columns defined on the Table subclass.
        '''

        df, name, extra_ok = getargs('df', 'name', 'extra_ok', kwargs)

        cls_cols = list([col['name'] for col in getattr(cls, '__columns__')])
        df_cols = list(df.columns)

        missing_columns = set(cls_cols) - set(df_cols)
        extra_columns = set(df_cols) - set(cls_cols)

        if extra_columns:
            raise ValueError(
                'unrecognized column(s) {} for table class {} (columns {})'.format(
                    extra_columns, cls.__name__, cls_cols
                )
            )

        use_index = False
        if len(missing_columns) == 1 and list(missing_columns)[0] == df.index.name:
            use_index = True

        elif missing_columns:
            raise ValueError(
                'missing column(s) {} for table class {} (columns {}, provided {})'.format(
                    missing_columns, cls.__name__, cls_cols, df_cols
                )
            )

        data = []
        for index, row in df.iterrows():
            if use_index:
                data.append([
                    row[colname] if colname != df.index.name else index
                    for colname in cls_cols
                ])
            else:
                data.append(tuple([row[colname] for colname in cls_cols]))

        if name is None:
            return cls(data=data)
        return cls(name=name, data=data)



import os
import tempfile
from copy import copy, deepcopy

from hdmf.build import (ObjectMapper, TypeMap, BuildManager)
from hdmf.container import (Container, Data)
from hdmf.spec import (GroupSpec, DatasetSpec, AttributeSpec, LinkSpec,
                       RefSpec, DtypeSpec, NamespaceCatalog, SpecCatalog,
                       SpecNamespace, NamespaceBuilder)
from hdmf.spec.spec import (ZERO_OR_MANY, ONE_OR_MANY, ZERO_OR_ONE)
from hdmf.utils import (docval, getargs, get_docval)

CORE_NAMESPACE = 'test_core'


class CacheSpecTestHelper(object):

    @staticmethod
    def get_types(catalog):
        types = set()
        for ns_name in catalog.namespaces:
            ns = catalog.get_namespace(ns_name)
            for source in ns['schema']:
                types.update(catalog.get_types(source['source']))
        return types


def get_temp_filepath():
    # On Windows, h5py cannot truncate an open file in write mode.
    # The temp file will be closed before h5py truncates it and will be removed during the tearDown step.
    temp_file = tempfile.NamedTemporaryFile()
    temp_file.close()
    return temp_file.name


############################################
#  Foo example data containers and specs
###########################################
class Foo(Container):

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this Foo'},
            {'name': 'my_data', 'type': ('array_data', 'data'), 'doc': 'some data'},
            {'name': 'attr1', 'type': str, 'doc': 'an attribute'},
            {'name': 'attr2', 'type': int, 'doc': 'another attribute'},
            {'name': 'attr3', 'type': float, 'doc': 'a third attribute', 'default': 3.14})
    def __init__(self, **kwargs):
        name, my_data, attr1, attr2, attr3 = getargs('name', 'my_data', 'attr1', 'attr2', 'attr3', kwargs)
        super().__init__(name=name)
        self.__data = my_data
        self.__attr1 = attr1
        self.__attr2 = attr2
        self.__attr3 = attr3

    def __eq__(self, other):
        attrs = ('name', 'my_data', 'attr1', 'attr2', 'attr3')
        return all(getattr(self, a) == getattr(other, a) for a in attrs)

    def __str__(self):
        attrs = ('name', 'my_data', 'attr1', 'attr2', 'attr3')
        return '<' + ','.join('%s=%s' % (a, getattr(self, a)) for a in attrs) + '>'

    @property
    def my_data(self):
        return self.__data

    @property
    def attr1(self):
        return self.__attr1

    @property
    def attr2(self):
        return self.__attr2

    @property
    def attr3(self):
        return self.__attr3

    def __hash__(self):
        return hash(self.name)


class FooBucket(Container):

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this bucket'},
            {'name': 'foos', 'type': list, 'doc': 'the Foo objects in this bucket', 'default': list()})
    def __init__(self, **kwargs):
        name, foos = getargs('name', 'foos', kwargs)
        super().__init__(name=name)
        self.__foos = {f.name: f for f in foos}  # note: collections of groups are unordered in HDF5
        for f in foos:
            f.parent = self

    def __eq__(self, other):
        return self.name == other.name and self.foos == other.foos

    def __str__(self):
        return 'name=%s, foos=%s' % (self.name, self.foos)

    @property
    def foos(self):
        return self.__foos

    def remove_foo(self, foo_name):
        foo = self.__foos.pop(foo_name)
        if foo.parent is self:
            self._remove_child(foo)
        return foo


class FooFile(Container):
    """
    NOTE: if the ROOT_NAME for the backend is not 'root' then we must set FooFile.ROOT_NAME before use
          and should be reset to 'root' when use is finished to avoid potential cross-talk between tests.
    """

    ROOT_NAME = 'root'  # For HDF5 and Zarr this is the root. It should be set before use if different for the backend.

    @docval({'name': 'buckets', 'type': list, 'doc': 'the FooBuckets in this file', 'default': list()},
            {'name': 'foo_link', 'type': Foo, 'doc': 'an optional linked Foo', 'default': None},
            {'name': 'foofile_data', 'type': 'array_data', 'doc': 'an optional dataset', 'default': None},
            {'name': 'foo_ref_attr', 'type': Foo, 'doc': 'a reference Foo', 'default': None},
            )
    def __init__(self, **kwargs):
        buckets, foo_link, foofile_data, foo_ref_attr = getargs('buckets', 'foo_link', 'foofile_data',
                                                                'foo_ref_attr', kwargs)
        super().__init__(name=self.ROOT_NAME)  # name is not used - FooFile should be the root container
        self.__buckets = {b.name: b for b in buckets}  # note: collections of groups are unordered in HDF5
        for f in buckets:
            f.parent = self
        self.__foo_link = foo_link
        self.__foofile_data = foofile_data
        self.__foo_ref_attr = foo_ref_attr

    def __eq__(self, other):
        return (self.buckets == other.buckets
                and self.foo_link == other.foo_link
                and self.foofile_data == other.foofile_data)

    def __str__(self):
        return ('buckets=%s, foo_link=%s, foofile_data=%s' % (self.buckets, self.foo_link, self.foofile_data))

    @property
    def buckets(self):
        return self.__buckets

    def add_bucket(self, bucket):
        self.__buckets[bucket.name] = bucket
        bucket.parent = self

    def remove_bucket(self, bucket_name):
        bucket = self.__buckets.pop(bucket_name)
        if bucket.parent is self:
            self._remove_child(bucket)
        return bucket

    @property
    def foo_link(self):
        return self.__foo_link

    @foo_link.setter
    def foo_link(self, value):
        if self.__foo_link is None:
            self.__foo_link = value
            self.set_modified(True)
        else:
            raise ValueError("can't reset foo_link attribute")

    @property
    def foofile_data(self):
        return self.__foofile_data

    @foofile_data.setter
    def foofile_data(self, value):
        if self.__foofile_data is None:
            self.__foofile_data = value
            self.set_modified(True)
        else:
            raise ValueError("can't reset foofile_data attribute")

    @property
    def foo_ref_attr(self):
        return self.__foo_ref_attr

    @foo_ref_attr.setter
    def foo_ref_attr(self, value):
        if self.__foo_ref_attr is None:
            self.__foo_ref_attr = value
            self.set_modified(True)
        else:
            raise ValueError("can't reset foo_ref_attr attribute")


def get_foo_buildmanager():
    """
    Get a BuildManager (and create all ObjectMappers) for a foofile
    :return:
    """

    foo_spec = GroupSpec('A test group specification with a data type',
                         data_type_def='Foo',
                         datasets=[DatasetSpec('an example dataset',
                                               'int',
                                               name='my_data',
                                               attributes=[AttributeSpec('attr2',
                                                                         'an example integer attribute',
                                                                         'int')])],
                         attributes=[AttributeSpec('attr1', 'an example string attribute', 'text'),
                                     AttributeSpec('attr3', 'an example float attribute', 'float')])

    tmp_spec = GroupSpec('A subgroup for Foos',
                         name='foo_holder',
                         groups=[GroupSpec('the Foos in this bucket', data_type_inc='Foo', quantity=ZERO_OR_MANY)])

    bucket_spec = GroupSpec('A test group specification for a data type containing data type',
                            data_type_def='FooBucket',
                            groups=[tmp_spec])

    class FooMapper(ObjectMapper):
        def __init__(self, spec):
            super().__init__(spec)
            my_data_spec = spec.get_dataset('my_data')
            self.map_spec('attr2', my_data_spec.get_attribute('attr2'))

    class BucketMapper(ObjectMapper):
        def __init__(self, spec):
            super().__init__(spec)
            foo_holder_spec = spec.get_group('foo_holder')
            self.unmap(foo_holder_spec)
            foo_spec = foo_holder_spec.get_data_type('Foo')
            self.map_spec('foos', foo_spec)

    file_links_spec = GroupSpec('Foo link group',
                                name='links',
                                links=[LinkSpec('Foo link',
                                                name='foo_link',
                                                target_type='Foo',
                                                quantity=ZERO_OR_ONE)]
                                )

    file_spec = GroupSpec("A file of Foos contained in FooBuckets",
                          data_type_def='FooFile',
                          groups=[GroupSpec('Holds the FooBuckets',
                                            name='buckets',
                                            groups=[GroupSpec("One or more FooBuckets",
                                                              data_type_inc='FooBucket',
                                                              quantity=ZERO_OR_MANY)]),
                                  file_links_spec],
                          datasets=[DatasetSpec('Foo data',
                                                name='foofile_data',
                                                dtype='int',
                                                quantity=ZERO_OR_ONE)],
                          attributes=[AttributeSpec(doc='Foo ref attr',
                                                    name='foo_ref_attr',
                                                    dtype=RefSpec('Foo', 'object'),
                                                    required=False)],
                          )

    class FileMapper(ObjectMapper):
        def __init__(self, spec):
            super().__init__(spec)
            bucket_spec = spec.get_group('buckets').get_data_type('FooBucket')
            self.map_spec('buckets', bucket_spec)
            self.unmap(spec.get_group('links'))
            foo_link_spec = spec.get_group('links').get_link('foo_link')
            self.map_spec('foo_link', foo_link_spec)

    spec_catalog = SpecCatalog()
    spec_catalog.register_spec(foo_spec, 'test.yaml')
    spec_catalog.register_spec(bucket_spec, 'test.yaml')
    spec_catalog.register_spec(file_spec, 'test.yaml')
    namespace = SpecNamespace(
        'a test namespace',
        CORE_NAMESPACE,
        [{'source': 'test.yaml'}],
        version='0.1.0',
        catalog=spec_catalog)
    namespace_catalog = NamespaceCatalog()
    namespace_catalog.add_namespace(CORE_NAMESPACE, namespace)
    type_map = TypeMap(namespace_catalog)

    type_map.register_container_type(CORE_NAMESPACE, 'Foo', Foo)
    type_map.register_container_type(CORE_NAMESPACE, 'FooBucket', FooBucket)
    type_map.register_container_type(CORE_NAMESPACE, 'FooFile', FooFile)

    type_map.register_map(Foo, FooMapper)
    type_map.register_map(FooBucket, BucketMapper)
    type_map.register_map(FooFile, FileMapper)

    manager = BuildManager(type_map)
    return manager


############################################
#  Baz example data containers and specs
###########################################
class Baz(Container):

    pass


class BazData(Data):

    pass


class BazCpdData(Data):

    pass


class BazBucket(Container):

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this bucket'},
            {'name': 'bazs', 'type': list, 'doc': 'the Baz objects in this bucket'},
            {'name': 'baz_data', 'type': BazData, 'doc': 'dataset of Baz references', 'default': None},
            {'name': 'baz_cpd_data', 'type': BazCpdData, 'doc': 'dataset of Baz references', 'default': None})
    def __init__(self, **kwargs):
        name, bazs, baz_data, baz_cpd_data = getargs('name', 'bazs', 'baz_data', 'baz_cpd_data', kwargs)
        super().__init__(name=name)
        self.__bazs = {b.name: b for b in bazs}  # note: collections of groups are unordered in HDF5
        for b in bazs:
            b.parent = self
        self.__baz_data = baz_data
        if self.__baz_data is not None:
            self.__baz_data.parent = self
        self.__baz_cpd_data = baz_cpd_data
        if self.__baz_cpd_data is not None:
            self.__baz_cpd_data.parent = self

    @property
    def bazs(self):
        return self.__bazs

    @property
    def baz_data(self):
        return self.__baz_data

    @property
    def baz_cpd_data(self):
        return self.__baz_cpd_data

    def add_baz(self, baz):
        self.__bazs[baz.name] = baz
        baz.parent = self

    def remove_baz(self, baz_name):
        baz = self.__bazs.pop(baz_name)
        self._remove_child(baz)
        return baz


def get_baz_buildmanager():
    baz_spec = GroupSpec(
        doc='A test group specification with a data type',
        data_type_def='Baz',
    )

    baz_data_spec = DatasetSpec(
        doc='A test dataset of references specification with a data type',
        name='baz_data',
        data_type_def='BazData',
        dtype=RefSpec('Baz', 'object'),
        shape=[None],
    )

    baz_cpd_data_spec = DatasetSpec(
        doc='A test compound dataset with references specification with a data type',
        name='baz_cpd_data',
        data_type_def='BazCpdData',
        dtype=[DtypeSpec(name='part1', doc='doc', dtype='int'),
               DtypeSpec(name='part2', doc='doc', dtype=RefSpec('Baz', 'object'))],
        shape=[None],
    )

    baz_holder_spec = GroupSpec(
        doc='group of bazs',
        name='bazs',
        groups=[GroupSpec(doc='Baz', data_type_inc='Baz', quantity=ONE_OR_MANY)],
    )

    baz_bucket_spec = GroupSpec(
        doc='A test group specification for a data type containing data type',
        data_type_def='BazBucket',
        groups=[baz_holder_spec],
        datasets=[DatasetSpec(doc='doc', data_type_inc='BazData', quantity=ZERO_OR_ONE),
                  DatasetSpec(doc='doc', data_type_inc='BazCpdData', quantity=ZERO_OR_ONE)],
    )

    spec_catalog = SpecCatalog()
    spec_catalog.register_spec(baz_spec, 'test.yaml')
    spec_catalog.register_spec(baz_data_spec, 'test.yaml')
    spec_catalog.register_spec(baz_cpd_data_spec, 'test.yaml')
    spec_catalog.register_spec(baz_bucket_spec, 'test.yaml')

    namespace = SpecNamespace(
        'a test namespace',
        CORE_NAMESPACE,
        [{'source': 'test.yaml'}],
        version='0.1.0',
        catalog=spec_catalog)

    namespace_catalog = NamespaceCatalog()
    namespace_catalog.add_namespace(CORE_NAMESPACE, namespace)

    type_map = TypeMap(namespace_catalog)
    type_map.register_container_type(CORE_NAMESPACE, 'Baz', Baz)
    type_map.register_container_type(CORE_NAMESPACE, 'BazData', BazData)
    type_map.register_container_type(CORE_NAMESPACE, 'BazCpdData', BazCpdData)
    type_map.register_container_type(CORE_NAMESPACE, 'BazBucket', BazBucket)

    class BazBucketMapper(ObjectMapper):
        def __init__(self, spec):
            super().__init__(spec)
            baz_holder_spec = spec.get_group('bazs')
            self.unmap(baz_holder_spec)
            baz_spec = baz_holder_spec.get_data_type('Baz')
            self.map_spec('bazs', baz_spec)

    type_map.register_map(BazBucket, BazBucketMapper)

    manager = BuildManager(type_map)
    return manager


def create_test_type_map(specs, container_classes, mappers=None):
    """
    Create a TypeMap with the specs registered under a test namespace, and classes and mappers registered to type names.
    :param specs: list of specs
    :param container_classes: dict of type name to container class
    :param mappers: (optional) dict of type name to mapper class
    :return: the constructed TypeMap
    """
    spec_catalog = SpecCatalog()
    schema_file = 'test.yaml'
    for s in specs:
        spec_catalog.register_spec(s, schema_file)
    namespace = SpecNamespace(
        doc='a test namespace',
        name=CORE_NAMESPACE,
        schema=[{'source': schema_file}],
        version='0.1.0',
        catalog=spec_catalog
    )
    namespace_catalog = NamespaceCatalog()
    namespace_catalog.add_namespace(CORE_NAMESPACE, namespace)
    type_map = TypeMap(namespace_catalog)
    for type_name, container_cls in container_classes.items():
        type_map.register_container_type(CORE_NAMESPACE, type_name, container_cls)
    if mappers:
        for type_name, mapper_cls in mappers.items():
            container_cls = container_classes[type_name]
            type_map.register_map(container_cls, mapper_cls)
    return type_map


def create_load_namespace_yaml(namespace_name, specs, output_dir, incl_types, type_map):
    """
    Create a TypeMap with the specs loaded from YAML files and dependencies resolved.

    This writes namespaces and specs to YAML files, creates an empty TypeMap, and calls load_namespaces on the
    TypeMap, instead of manually creating a SpecCatalog, SpecNamespace, NamespaceCatalog and manually registering
    container types. Importantly, this process resolves dependencies across namespaces.

    :param namespace_name: Name of the new namespace.
    :param specs: List of specs of new data types to add.
    :param incl_types: Dict mapping included namespace name to list of data types to include or None to include all.
    :param type_map: The type map to load the namespace into.
    """
    ns_builder = NamespaceBuilder(
        name=namespace_name,
        doc='a test namespace',
        version='0.1.0',
    )
    ns_filename = ns_builder.name + '.namespace.yaml'
    ext_filename = ns_builder.name + '.extensions.yaml'

    for ns, types in incl_types.items():
        if types is None:  # include all types
            ns_builder.include_namespace(ns)
        else:
            for dt in types:
                ns_builder.include_type(dt, namespace=ns)

    for data_type in specs:
        ns_builder.add_spec(ext_filename, data_type)

    ns_builder.export(ns_filename, outdir=output_dir)
    ns_path = os.path.join(output_dir, ns_filename)
    type_map.load_namespaces(ns_path)


# ##### custom spec classes #####

def swap_inc_def(cls, custom_cls):
    args = get_docval(cls.__init__)
    ret = list()
    for arg in args:
        if arg['name'] == 'data_type_def':
            ret.append({'name': 'my_data_type_def', 'type': str,
                        'doc': 'the NWB data type this spec defines', 'default': None})
        elif arg['name'] == 'data_type_inc':
            ret.append({'name': 'my_data_type_inc', 'type': (custom_cls, str),
                        'doc': 'the NWB data type this spec includes', 'default': None})
        else:
            ret.append(copy(arg))
    return ret


class BaseStorageOverride:
    __type_key = 'my_data_type'
    __inc_key = 'my_data_type_inc'
    __def_key = 'my_data_type_def'

    @classmethod
    def type_key(cls):
        ''' Get the key used to store data type on an instance'''
        return cls.__type_key

    @classmethod
    def inc_key(cls):
        ''' Get the key used to define a data_type include.'''
        return cls.__inc_key

    @classmethod
    def def_key(cls):
        ''' Get the key used to define a data_type definition.'''
        return cls.__def_key

    @classmethod
    def build_const_args(cls, spec_dict):
        """Extend base functionality to remap data_type_def and data_type_inc keys"""
        spec_dict = copy(spec_dict)
        proxy = super(BaseStorageOverride, cls)
        if proxy.inc_key() in spec_dict:
            spec_dict[cls.inc_key()] = spec_dict.pop(proxy.inc_key())
        if proxy.def_key() in spec_dict:
            spec_dict[cls.def_key()] = spec_dict.pop(proxy.def_key())
        ret = proxy.build_const_args(spec_dict)
        return ret

    @classmethod
    def _translate_kwargs(cls, kwargs):
        """Swap mydata_type_def and mydata_type_inc for data_type_def and data_type_inc, respectively"""
        proxy = super(BaseStorageOverride, cls)
        kwargs[proxy.def_key()] = kwargs.pop(cls.def_key())
        kwargs[proxy.inc_key()] = kwargs.pop(cls.inc_key())
        return kwargs


class CustomGroupSpec(BaseStorageOverride, GroupSpec):

    @docval(*deepcopy(swap_inc_def(GroupSpec, 'CustomGroupSpec')))
    def __init__(self, **kwargs):
        kwargs = self._translate_kwargs(kwargs)
        super().__init__(**kwargs)

    @classmethod
    def dataset_spec_cls(cls):
        return CustomDatasetSpec

    @docval(*deepcopy(swap_inc_def(GroupSpec, 'CustomGroupSpec')))
    def add_group(self, **kwargs):
        spec = CustomGroupSpec(**kwargs)
        self.set_group(spec)
        return spec

    @docval(*deepcopy(swap_inc_def(DatasetSpec, 'CustomDatasetSpec')))
    def add_dataset(self, **kwargs):
        ''' Add a new specification for a subgroup to this group specification '''
        spec = CustomDatasetSpec(**kwargs)
        self.set_dataset(spec)
        return spec


class CustomDatasetSpec(BaseStorageOverride, DatasetSpec):

    @docval(*deepcopy(swap_inc_def(DatasetSpec, 'CustomDatasetSpec')))
    def __init__(self, **kwargs):
        kwargs = self._translate_kwargs(kwargs)
        super().__init__(**kwargs)


class CustomSpecNamespace(SpecNamespace):
    __types_key = 'my_data_types'

    @classmethod
    def types_key(cls):
        return cls.__types_key



"""
Collection of Container classes for interacting with  aligned and hierarchical dynamic tables
"""
from collections import OrderedDict

import numpy as np
import pandas as pd

from . import register_class
from .table import DynamicTable
from ..utils import docval, getargs, popargs, get_docval, AllowPositional


@register_class('AlignedDynamicTable')
class AlignedDynamicTable(DynamicTable):
    """
    DynamicTable container that supports storing a collection of subtables. Each sub-table is a
    DynamicTable itself that is aligned with the main table by row index. I.e., all
    DynamicTables stored in this group MUST have the same number of rows. This type effectively
    defines a 2-level table in which the main data is stored in the main table implemented by this type
    and additional columns of the table are grouped into categories, with each category being'
    represented by a separate DynamicTable stored within the group.

    NOTE: To remain compatible with DynamicTable, the attribute colnames represents only the
          columns of the main table (not including the category tables). To get the full list of
          column names, use the get_colnames() function instead.
    """
    __fields__ = ({'name': 'category_tables', 'child': True}, )

    @docval(*get_docval(DynamicTable.__init__),
            {'name': 'category_tables', 'type': list,
             'doc': 'List of DynamicTables to be added to the container. NOTE: Only regular '
                    'DynamicTables are allowed. Using AlignedDynamicTable as a category for '
                    'AlignedDynamicTable is currently not supported.', 'default': None},
            {'name': 'categories', 'type': 'array_data',
             'doc': 'List of names with the ordering of category tables', 'default': None},
            allow_positional=AllowPositional.WARNING)
    def __init__(self, **kwargs):  # noqa: C901
        in_category_tables = popargs('category_tables', kwargs)
        in_categories = popargs('categories', kwargs)
        if in_category_tables is not None:
            # Error check to make sure that all category_table are regular DynamicTable
            for i, v in enumerate(in_category_tables):
                if not isinstance(v, DynamicTable):
                    raise ValueError("Category table with index %i is not a DynamicTable" % i)
                if isinstance(v, AlignedDynamicTable):
                    raise ValueError("Category table with index %i is an AlignedDynamicTable. "
                                     "Nesting of AlignedDynamicTable is currently not supported." % i)
        # set in_categories from the in_category_tables if it is empy
        if in_categories is None and in_category_tables is not None:
            in_categories = [tab.name for tab in in_category_tables]
        # check that if categories is given that we also have category_tables
        if in_categories is not None and in_category_tables is None:
            raise ValueError("Categories provided but no category_tables given")
        # at this point both in_categories and in_category_tables should either both be None or both be a list
        if in_categories is not None:
            if len(in_categories) != len(in_category_tables):
                raise ValueError("%s category_tables given but %s categories specified" %
                                 (len(in_category_tables), len(in_categories)))
        # Initialize the main dynamic table
        super().__init__(**kwargs)
        # Create and set all sub-categories
        dts = OrderedDict()
        # Add the custom categories given as inputs
        if in_category_tables is not None:
            # We may need to resize our main table when adding categories as the user may not have set ids
            if len(in_category_tables) > 0:
                # We have categories to process
                if len(self.id) == 0:
                    # The user did not initialize our main table id's nor set columns for our main table
                    for i in range(len(in_category_tables[0])):
                        self.id.append(i)
            # Add the user-provided categories in the correct order as described by the categories
            # This is necessary, because we do not store the categories explicitly but we maintain them
            # as the order of our self.category_tables. In this makes sure look-ups are consistent.
            lookup_index = OrderedDict([(k, -1) for k in in_categories])
            for i, v in enumerate(in_category_tables):
                # Error check that the name of the table is in our categories list
                if v.name not in lookup_index:
                    raise ValueError("DynamicTable %s does not appear in categories %s" % (v.name, str(in_categories)))
                # Error check to make sure no two tables with the same name are given
                if lookup_index[v.name] >= 0:
                    raise ValueError("Duplicate table name %s found in input dynamic_tables" % v.name)
                lookup_index[v.name] = i
            for table_name, tabel_index in lookup_index.items():
                # This error case should not be able to occur since the length of the in_categories and
                # in_category_tables must match and we made sure that each DynamicTable we added had its
                # name in the in_categories list. We, therefore, exclude this check from coverage testing
                # but we leave it in just as a backup trigger in case something unexpected happens
                if tabel_index < 0:  # pragma: no cover
                    raise ValueError("DynamicTable %s listed in categories but does not appear in category_tables" %
                                     table_name)  # pragma: no cover
                # Test that all category tables have the correct number of rows
                category = in_category_tables[tabel_index]
                if len(category) != len(self):
                    raise ValueError('Category DynamicTable %s does not align, it has %i rows expected %i' %
                                     (category.name, len(category), len(self)))
                # Add the category table to our category_tables.
                dts[category.name] = category
        # Set the self.category_tables attribute, which will set the parent/child relationships for the category_tables
        self.category_tables = dts

    def __contains__(self, val):
        """
        Check if the given value (i.e., column) exists in this table

        :param val: If val is a string then check if the given category exists. If val is a tuple
        of two strings (category, colname) then check for the given category if the given colname exists.
        """
        if isinstance(val, str):
            return val in self.category_tables or val in self.colnames
        elif isinstance(val, tuple):
            if len(val) != 2:
                raise ValueError("Expected tuple of strings of length 2 got tuple of length %i" % len(val))
            return val[1] in self.get_category(val[0])
        else:
            return False

    @property
    def categories(self):
        """
        Get the list of names the categories

        Short-hand for list(self.category_tables.keys())

        :raises: KeyError if the given name is not in self.category_tables
        """
        return list(self.category_tables.keys())

    @docval({'name': 'category', 'type': DynamicTable, 'doc': 'Add a new DynamicTable category'},)
    def add_category(self, **kwargs):
        """
        Add a new DynamicTable to the AlignedDynamicTable to create a new category in the table.

        NOTE: The table must align with (i.e, have the same number of rows as) the main data table (and
        other category tables). I.e., if the AlignedDynamicTable is already populated with data
        then we have to populate the new category with the corresponding data before adding it.

        :raises: ValueError is raised if the input table does not have the same number of rows as the main table.
                 ValueError is raised if the table is an AlignedDynamicTable instead of regular DynamicTable.
        """
        category = getargs('category', kwargs)
        if len(category) != len(self):
            raise ValueError('New category DynamicTable does not align, it has %i rows expected %i' %
                             (len(category), len(self)))
        if category.name in self.category_tables:
            raise ValueError("Category %s already in the table" % category.name)
        if isinstance(category, AlignedDynamicTable):
            raise ValueError("Category is an AlignedDynamicTable. Nesting of AlignedDynamicTable "
                             "is currently not supported.")
        self.category_tables[category.name] = category
        category.parent = self

    @docval({'name': 'name', 'type': str, 'doc': 'Name of the category we want to retrieve', 'default': None})
    def get_category(self, **kwargs):
        name = popargs('name', kwargs)
        if name is None or (name not in self.category_tables and name == self.name):
            return self
        else:
            return self.category_tables[name]

    @docval(*get_docval(DynamicTable.add_column),
            {'name': 'category', 'type': str, 'doc': 'The category the column should be added to',
             'default': None})
    def add_column(self, **kwargs):
        """
        Add a column to the table

        :raises: KeyError if the category does not exist

        """
        category_name = popargs('category', kwargs)
        if category_name is None:
            # Add the column to our main table
            super().add_column(**kwargs)
        else:
            # Add the column to a sub-category table
            try:
                category = self.get_category(category_name)
            except KeyError:
                raise KeyError("Category %s not in table" % category_name)
            category.add_column(**kwargs)

    @docval({'name': 'data', 'type': dict, 'doc': 'the data to put in this row', 'default': None},
            {'name': 'id', 'type': int, 'doc': 'the ID for the row', 'default': None},
            {'name': 'enforce_unique_id', 'type': bool, 'doc': 'enforce that the id in the table must be unique',
             'default': False},
            allow_extra=True)
    def add_row(self, **kwargs):
        """
        We can either provide the row data as a single dict or by specifying a dict for each category
        """
        data, row_id, enforce_unique_id = popargs('data', 'id', 'enforce_unique_id', kwargs)
        data = data if data is not None else kwargs

        # extract the category data
        category_data = {k: data.pop(k) for k in self.categories if k in data}

        # Check that we have the approbriate categories provided
        missing_categories = set(self.categories) - set(list(category_data.keys()))
        if missing_categories:
            raise KeyError(
                '\n'.join([
                    'row data keys do not match available categories',
                    'missing {} category keys: {}'.format(len(missing_categories), missing_categories)
                ])
            )
        # Add the data to our main dynamic table
        data['id'] = row_id
        data['enforce_unique_id'] = enforce_unique_id
        super().add_row(**data)

        # Add the data to all out dynamic table categories
        for category, values in category_data.items():
            self.category_tables[category].add_row(**values)

    @docval({'name': 'include_category_tables', 'type': bool,
             'doc': "Ignore sub-category tables and just look at the main table", 'default': False},
            {'name': 'ignore_category_ids', 'type': bool,
             'doc': "Ignore id columns of sub-category tables", 'default': False})
    def get_colnames(self, **kwargs):
        """Get the full list of names of columns for this table

        :returns: List of tuples (str, str) where the first string is the name of the DynamicTable
                  that contains the column and the second string is the name of the column. If
                  include_category_tables is False, then a list of column names is returned.
        """
        if not getargs('include_category_tables', kwargs):
            return self.colnames
        else:
            ignore_category_ids = getargs('ignore_category_ids', kwargs)
            columns = [(self.name, c) for c in self.colnames]
            for category in self.category_tables.values():
                if not ignore_category_ids:
                    columns += [(category.name, 'id'), ]
                columns += [(category.name, c) for c in category.colnames]
            return columns

    @docval({'name': 'ignore_category_ids', 'type': bool,
             'doc': "Ignore id columns of sub-category tables", 'default': False})
    def to_dataframe(self, **kwargs):
        """Convert the collection of tables to a single pandas DataFrame"""
        dfs = [super().to_dataframe().reset_index(), ]
        if getargs('ignore_category_ids', kwargs):
            dfs += [category.to_dataframe() for category in self.category_tables.values()]
        else:
            dfs += [category.to_dataframe().reset_index() for category in self.category_tables.values()]
        names = [self.name, ] + list(self.category_tables.keys())
        res = pd.concat(dfs, axis=1, keys=names)
        res.set_index((self.name, 'id'), drop=True, inplace=True)
        return res

    def __getitem__(self, item):
        """
        Called to implement standard array slicing syntax.

        Same as ``self.get(item)``. See :py:meth:`~hdmf.common.alignedtable.AlignedDynamicTable.get` for details.
        """
        return self.get(item)

    def get(self, item, **kwargs):
        """
        Access elements (rows, columns, category tables etc.) from the table. Instead of calling
        this function directly, the class also implements standard array slicing syntax
        via :py:meth:`~hdmf.common.alignedtable.AlignedDynamicTable.__getitem__`
        (which calls this function). For example, instead of calling
        ``self.get(item=slice(2,5))`` we may use the often more convenient form of ``self[2:5]`` instead.

        :param item: Selection defining the items of interest. This may be either a:

        * **int, list, array, slice** : Return one or multiple row of the table as a pandas.DataFrame. For example:
              * ``self[0]`` : Select the first row of the table
              * ``self[[0,3]]`` : Select the first and fourth row of the table
              * ``self[1:4]`` : Select the rows with index 1,2,3 from the table

        * **string** : Return a column from the main table or a category table. For example:
              * ``self['column']`` : Return the column from the main table.
              * ``self['my_category']`` : Returns a DataFrame of the ``my_category`` category table.
                This is a shorthand for ``self.get_category('my_category').to_dataframe()``.

        * **tuple**: Get a column, row, or cell from a particular category table. The tuple is expected to
          consist of the following elements:

              * ``category``: string with the name of the category. To select from the main
                table use ``self.name`` or ``None``.
              * ``column``: string with the name of the column, and
              * ``row``: integer index of the row.

          The tuple itself then may take the following forms:

              * Select a single column from a table via:
                   * ``self[category, column]``
              * Select a single full row of a given category table via:
                   * ``self[row, category]`` (recommended, for consistency with DynamicTable)
                   * ``self[category, row]``
              * Select a single cell via:
                   * ``self[row, (category, column)]`` (recommended, for consistency with DynamicTable)
                   * ``self[row, category, column]``
                   * ``self[category, column, row]``

        :returns: Depending on the type of selection the function returns a:

            * **pandas.DataFrame**: when retrieving a row or category table
            * **array** : when retrieving a single column
            * **single value** : when retrieving a single cell. The data type and shape will depend on the
              data type and shape of the cell/column.
        """
        if isinstance(item, (int, list, np.ndarray, slice)):
            # get a single full row from all tables
            dfs = ([super().get(item, **kwargs).reset_index(), ] +
                   [category[item].reset_index() for category in self.category_tables.values()])
            names = [self.name, ] + list(self.category_tables.keys())
            res = pd.concat(dfs, axis=1, keys=names)
            res.set_index((self.name, 'id'), drop=True, inplace=True)
            return res
        elif isinstance(item, str) or item is None:
            if item in self.colnames:
                # get a specific column
                return super().get(item, **kwargs)
            else:
                # get a single category
                return self.get_category(item).to_dataframe()
        elif isinstance(item, tuple):
            if len(item) == 2:
                # DynamicTable allows selection of cells via the syntax [int, str], i.e,. [row_index, columnname]
                # We support this syntax here as well with the additional caveat that in AlignedDynamicTable
                # columns are identified by tuples of strings. As such [int, str] refers not to a cell but
                # a single row in a particular category table (i.e., [row_index, category]). To select a cell
                # the second part of the item then is a tuple of strings, i.e., [row_index, (category, column)]
                if isinstance(item[0], (int, np.integer)):
                    # Select a single cell or row of a sub-table based on row-index(item[0])
                    # and the category (if item[1] is a string) or column (if item[1] is a tuple of (category, column)
                    re = self[item[0]][item[1]]
                    # re is a pandas.Series or pandas.Dataframe. If we selected a single cell
                    # (i.e., item[2] was a tuple defining a particular column) then return the value of the cell
                    if re.size == 1:
                        re = re.values[0]
                        # If we selected a single cell from a ragged column then we need to change the list to a tuple
                        if isinstance(re, list):
                            re = tuple(re)
                    # We selected a row of a whole table (i.e., item[2] identified only the category table,
                    # but not a particular column).
                    # Change the result from a pandas.Series to a pandas.DataFrame for consistency with DynamicTable
                    if isinstance(re, pd.Series):
                        re = re.to_frame()
                    return re
                else:
                    return self.get_category(item[0])[item[1]]
            elif len(item) == 3:
                if isinstance(item[0], (int, np.integer)):
                    return self.get_category(item[1])[item[2]][item[0]]
                else:
                    return self.get_category(item[0])[item[1]][item[2]]
            else:
                raise ValueError("Expected tuple of length 2 of the form [category, column], [row, category], "
                                 "[row, (category, column)] or a tuple of length 3 of the form "
                                 "[category, column, row], [row, category, column]")

    @docval({'name': 'ignore_category_tables', 'type': bool,
             'doc': "Ignore the category tables and only check in the main table columns", 'default': False},
            allow_extra=False)
    def has_foreign_columns(self, **kwargs):
        """
        Does the table contain DynamicTableRegion columns

        :returns: True if the table or any of the category tables contains a DynamicTableRegion column, else False
        """
        ignore_category_tables = getargs('ignore_category_tables', kwargs)
        if super().has_foreign_columns():
            return True
        if not ignore_category_tables:
            for table in self.category_tables.values():
                if table.has_foreign_columns():
                    return True
        return False

    @docval({'name': 'ignore_category_tables', 'type': bool,
             'doc': "Ignore the category tables and only check in the main table columns", 'default': False},
            allow_extra=False)
    def get_foreign_columns(self, **kwargs):
        """
        Determine the names of all columns that link to another DynamicTable, i.e.,
        find all DynamicTableRegion type columns. Similar to a foreign key in a
        database, a DynamicTableRegion column references elements in another table.

        :returns: List of tuples (str, str) where the first string is the name of the
                  category table (or None if the column is in the main table) and the
                  second string is the column name.
        """
        ignore_category_tables = getargs('ignore_category_tables', kwargs)
        col_names = [(None, col_name) for col_name in super().get_foreign_columns()]
        if not ignore_category_tables:
            for table in self.category_tables.values():
                col_names += [(table.name, col_name) for col_name in table.get_foreign_columns()]
        return col_names

    @docval(*get_docval(DynamicTable.get_linked_tables),
            {'name': 'ignore_category_tables', 'type': bool,
             'doc': "Ignore the category tables and only check in the main table columns", 'default': False},
            allow_extra=False)
    def get_linked_tables(self, **kwargs):
        """
        Get a list of the full list of all tables that are being linked to directly or indirectly
        from this table via foreign DynamicTableColumns included in this table or in any table that
        can be reached through DynamicTableRegion columns


        Returns: List of dicts with the following keys:
                * 'source_table' : The source table containing the DynamicTableRegion column
                * 'source_column' : The relevant DynamicTableRegion column in the 'source_table'
                * 'target_table' : The target DynamicTable; same as source_column.table.

        """
        ignore_category_tables = getargs('ignore_category_tables', kwargs)
        other_tables = None if ignore_category_tables else list(self.category_tables.values())
        return super().get_linked_tables(other_tables=other_tables)



import numpy as np
import os
import shutil
import tempfile

from hdmf.build import TypeMap, CustomClassGenerator
from hdmf.build.classgenerator import ClassGenerator, MCIClassGenerator
from hdmf.container import Container, Data, MultiContainerInterface, AbstractContainer
from hdmf.spec import GroupSpec, AttributeSpec, DatasetSpec, SpecCatalog, SpecNamespace, NamespaceCatalog, LinkSpec
from hdmf.testing import TestCase
from hdmf.utils import get_docval, docval

from .test_io_map import Bar
from tests.unit.utils import CORE_NAMESPACE, create_test_type_map, create_load_namespace_yaml


class TestClassGenerator(TestCase):

    def test_register_generator(self):
        """Test TypeMap.register_generator and ClassGenerator.register_generator."""

        class MyClassGenerator(CustomClassGenerator):

            @classmethod
            def apply_generator_to_field(cls, field_spec, bases, type_map):
                return True

            @classmethod
            def process_field_spec(cls, classdict, docval_args, parent_cls, attr_name, not_inherited_fields, type_map,
                                   spec):
                # append attr_name to classdict['__custom_fields__'] list
                classdict.setdefault('process_field_spec', list()).append(attr_name)

            @classmethod
            def post_process(cls, classdict, bases, docval_args, spec):
                classdict['post_process'] = True

        spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='Baz',
            attributes=[
                AttributeSpec(name='attr1', doc='a string attribute', dtype='text')
            ]
        )

        spec_catalog = SpecCatalog()
        spec_catalog.register_spec(spec, 'test.yaml')
        namespace = SpecNamespace(
            doc='a test namespace',
            name=CORE_NAMESPACE,
            schema=[{'source': 'test.yaml'}],
            version='0.1.0',
            catalog=spec_catalog
        )
        namespace_catalog = NamespaceCatalog()
        namespace_catalog.add_namespace(CORE_NAMESPACE, namespace)
        type_map = TypeMap(namespace_catalog)
        type_map.register_generator(MyClassGenerator)
        cls = type_map.get_dt_container_cls('Baz', CORE_NAMESPACE)

        self.assertEqual(cls.process_field_spec, ['attr1'])
        self.assertTrue(cls.post_process)

    def test_bad_generator(self):
        """Test that register_generator raises an error if the generator is not an instance of CustomClassGenerator."""

        class NotACustomClassGenerator:
            pass

        type_map = TypeMap()

        msg = 'Generator <.*> must be a subclass of CustomClassGenerator.'
        with self.assertRaisesRegex(ValueError, msg):
            type_map.register_generator(NotACustomClassGenerator)

    def test_no_generators(self):
        """Test that a ClassGenerator without registered generators does nothing."""
        cg = ClassGenerator()
        spec = GroupSpec(doc='A test group spec with a data type', data_type_def='Baz')
        cls = cg.generate_class(data_type='Baz', spec=spec, parent_cls=Container, attr_names={}, type_map=TypeMap())
        self.assertEqual(cls.__mro__, (cls, Container, AbstractContainer, object))
        self.assertTrue(hasattr(cls, '__init__'))


class TestDynamicContainer(TestCase):

    def setUp(self):
        self.bar_spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='Bar',
            datasets=[
                DatasetSpec(
                    doc='a dataset',
                    dtype='int',
                    name='data',
                    attributes=[AttributeSpec(name='attr2', doc='an integer attribute', dtype='int')]
                )
            ],
            attributes=[AttributeSpec(name='attr1', doc='a string attribute', dtype='text')])
        specs = [self.bar_spec]
        containers = {'Bar': Bar}
        self.type_map = create_test_type_map(specs, containers)
        self.spec_catalog = self.type_map.namespace_catalog.get_namespace(CORE_NAMESPACE).catalog

    def test_dynamic_container_creation(self):
        baz_spec = GroupSpec('A test extension with no Container class',
                             data_type_def='Baz', data_type_inc=self.bar_spec,
                             attributes=[AttributeSpec('attr3', 'a float attribute', 'float'),
                                         AttributeSpec('attr4', 'another float attribute', 'float')])
        self.spec_catalog.register_spec(baz_spec, 'extension.yaml')
        cls = self.type_map.get_dt_container_cls('Baz', CORE_NAMESPACE)
        expected_args = {'name', 'data', 'attr1', 'attr2', 'attr3', 'attr4'}
        received_args = set()
        for x in get_docval(cls.__init__):
            if x['name'] != 'foo':
                received_args.add(x['name'])
                with self.subTest(name=x['name']):
                    self.assertNotIn('default', x)
        self.assertSetEqual(expected_args, received_args)
        self.assertEqual(cls.__name__, 'Baz')
        self.assertTrue(issubclass(cls, Bar))

    def test_dynamic_container_default_name(self):
        baz_spec = GroupSpec('doc', default_name='bingo', data_type_def='Baz',
                             attributes=[AttributeSpec('attr4', 'another float attribute', 'float')])
        self.spec_catalog.register_spec(baz_spec, 'extension.yaml')
        cls = self.type_map.get_dt_container_cls('Baz', CORE_NAMESPACE)
        inst = cls(attr4=10.)
        self.assertEqual(inst.name, 'bingo')

    def test_dynamic_container_creation_defaults(self):
        baz_spec = GroupSpec('A test extension with no Container class',
                             data_type_def='Baz', data_type_inc=self.bar_spec,
                             attributes=[AttributeSpec('attr3', 'a float attribute', 'float'),
                                         AttributeSpec('attr4', 'another float attribute', 'float')])
        self.spec_catalog.register_spec(baz_spec, 'extension.yaml')
        cls = self.type_map.get_dt_container_cls('Baz', CORE_NAMESPACE)
        expected_args = {'name', 'data', 'attr1', 'attr2', 'attr3', 'attr4', 'foo'}
        received_args = set(map(lambda x: x['name'], get_docval(cls.__init__)))
        self.assertSetEqual(expected_args, received_args)
        self.assertEqual(cls.__name__, 'Baz')
        self.assertTrue(issubclass(cls, Bar))

    def test_dynamic_container_constructor(self):
        baz_spec = GroupSpec('A test extension with no Container class',
                             data_type_def='Baz', data_type_inc=self.bar_spec,
                             attributes=[AttributeSpec('attr3', 'a float attribute', 'float'),
                                         AttributeSpec('attr4', 'another float attribute', 'float')])
        self.spec_catalog.register_spec(baz_spec, 'extension.yaml')
        cls = self.type_map.get_dt_container_cls('Baz', CORE_NAMESPACE)
        # TODO: test that constructor works!
        inst = cls(name='My Baz', data=[1, 2, 3, 4], attr1='string attribute', attr2=1000, attr3=98.6, attr4=1.0)
        self.assertEqual(inst.name, 'My Baz')
        self.assertEqual(inst.data, [1, 2, 3, 4])
        self.assertEqual(inst.attr1, 'string attribute')
        self.assertEqual(inst.attr2, 1000)
        self.assertEqual(inst.attr3, 98.6)
        self.assertEqual(inst.attr4, 1.0)

    def test_dynamic_container_constructor_name(self):
        # name is specified in spec and cannot be changed
        baz_spec = GroupSpec('A test extension with no Container class',
                             data_type_def='Baz', data_type_inc=self.bar_spec,
                             name='A fixed name',
                             attributes=[AttributeSpec('attr3', 'a float attribute', 'float'),
                                         AttributeSpec('attr4', 'another float attribute', 'float')])
        self.spec_catalog.register_spec(baz_spec, 'extension.yaml')
        cls = self.type_map.get_dt_container_cls('Baz', CORE_NAMESPACE)

        with self.assertRaises(TypeError):
            inst = cls(name='My Baz', data=[1, 2, 3, 4], attr1='string attribute', attr2=1000, attr3=98.6, attr4=1.0)

        inst = cls(data=[1, 2, 3, 4], attr1='string attribute', attr2=1000, attr3=98.6, attr4=1.0)
        self.assertEqual(inst.name, 'A fixed name')
        self.assertEqual(inst.data, [1, 2, 3, 4])
        self.assertEqual(inst.attr1, 'string attribute')
        self.assertEqual(inst.attr2, 1000)
        self.assertEqual(inst.attr3, 98.6)
        self.assertEqual(inst.attr4, 1.0)

    def test_dynamic_container_constructor_name_default_name(self):
        # if both name and default_name are specified, name should be used
        with self.assertWarns(Warning):
            baz_spec = GroupSpec('A test extension with no Container class',
                                 data_type_def='Baz', data_type_inc=self.bar_spec,
                                 name='A fixed name',
                                 default_name='A default name',
                                 attributes=[AttributeSpec('attr3', 'a float attribute', 'float'),
                                             AttributeSpec('attr4', 'another float attribute', 'float')])
            self.spec_catalog.register_spec(baz_spec, 'extension.yaml')
            cls = self.type_map.get_dt_container_cls('Baz', CORE_NAMESPACE)

            inst = cls(data=[1, 2, 3, 4], attr1='string attribute', attr2=1000, attr3=98.6, attr4=1.0)
            self.assertEqual(inst.name, 'A fixed name')

    def test_dynamic_container_composition(self):
        baz_spec2 = GroupSpec('A composition inside', data_type_def='Baz2',
                              data_type_inc=self.bar_spec,
                              attributes=[
                                  AttributeSpec('attr3', 'a float attribute', 'float'),
                                  AttributeSpec('attr4', 'another float attribute', 'float')])

        baz_spec1 = GroupSpec('A composition test outside', data_type_def='Baz1', data_type_inc=self.bar_spec,
                              attributes=[AttributeSpec('attr3', 'a float attribute', 'float'),
                                          AttributeSpec('attr4', 'another float attribute', 'float')],
                              groups=[GroupSpec('A composition inside', data_type_inc='Baz2')])
        self.spec_catalog.register_spec(baz_spec1, 'extension.yaml')
        self.spec_catalog.register_spec(baz_spec2, 'extension.yaml')
        Baz2 = self.type_map.get_dt_container_cls('Baz2', CORE_NAMESPACE)
        Baz1 = self.type_map.get_dt_container_cls('Baz1', CORE_NAMESPACE)
        Baz1(name='My Baz', data=[1, 2, 3, 4], attr1='string attribute', attr2=1000, attr3=98.6, attr4=1.0,
             baz2=Baz2(name='My Baz', data=[1, 2, 3, 4], attr1='string attribute', attr2=1000, attr3=98.6, attr4=1.0))

        Bar = self.type_map.get_dt_container_cls('Bar', CORE_NAMESPACE)
        bar = Bar(name='My Bar', data=[1, 2, 3, 4], attr1='string attribute', attr2=1000)

        with self.assertRaises(TypeError):
            Baz1(name='My Baz', data=[1, 2, 3, 4], attr1='string attribute', attr2=1000, attr3=98.6, attr4=1.0,
                 baz2=bar)

    def test_dynamic_container_composition_reverse_order(self):
        baz_spec2 = GroupSpec('A composition inside', data_type_def='Baz2',
                              data_type_inc=self.bar_spec,
                              attributes=[
                                  AttributeSpec('attr3', 'a float attribute', 'float'),
                                  AttributeSpec('attr4', 'another float attribute', 'float')])

        baz_spec1 = GroupSpec('A composition test outside', data_type_def='Baz1', data_type_inc=self.bar_spec,
                              attributes=[AttributeSpec('attr3', 'a float attribute', 'float'),
                                          AttributeSpec('attr4', 'another float attribute', 'float')],
                              groups=[GroupSpec('A composition inside', data_type_inc='Baz2')])
        self.spec_catalog.register_spec(baz_spec1, 'extension.yaml')
        self.spec_catalog.register_spec(baz_spec2, 'extension.yaml')
        Baz1 = self.type_map.get_dt_container_cls('Baz1', CORE_NAMESPACE)
        Baz2 = self.type_map.get_dt_container_cls('Baz2', CORE_NAMESPACE)
        Baz1(name='My Baz', data=[1, 2, 3, 4], attr1='string attribute', attr2=1000, attr3=98.6, attr4=1.0,
             baz2=Baz2(name='My Baz', data=[1, 2, 3, 4], attr1='string attribute', attr2=1000, attr3=98.6, attr4=1.0))

        Bar = self.type_map.get_dt_container_cls('Bar', CORE_NAMESPACE)
        bar = Bar(name='My Bar', data=[1, 2, 3, 4], attr1='string attribute', attr2=1000)

        with self.assertRaises(TypeError):
            Baz1(name='My Baz', data=[1, 2, 3, 4], attr1='string attribute',
                 attr2=1000, attr3=98.6, attr4=1.0, baz2=bar)

    def test_dynamic_container_composition_missing_type(self):
        baz_spec1 = GroupSpec('A composition test outside', data_type_def='Baz1', data_type_inc=self.bar_spec,
                              attributes=[AttributeSpec('attr3', 'a float attribute', 'float'),
                                          AttributeSpec('attr4', 'another float attribute', 'float')],
                              groups=[GroupSpec('A composition inside', data_type_inc='Baz2')])
        self.spec_catalog.register_spec(baz_spec1, 'extension.yaml')

        msg = "No specification for 'Baz2' in namespace 'test_core'"
        with self.assertRaisesWith(ValueError, msg):
            self.type_map.get_dt_container_cls('Baz1', CORE_NAMESPACE)

    def test_dynamic_container_fixed_name(self):
        """Test that dynamic class generation for an extended type with a fixed name works."""
        baz_spec = GroupSpec('A test extension with no Container class',
                             data_type_def='Baz', data_type_inc=self.bar_spec, name='Baz')
        self.spec_catalog.register_spec(baz_spec, 'extension.yaml')
        Baz = self.type_map.get_dt_container_cls('Baz', CORE_NAMESPACE)
        obj = Baz(data=[1, 2, 3, 4], attr1='string attribute', attr2=1000)
        self.assertEqual(obj.name, 'Baz')

    def test_dynamic_container_super_init_fixed_value(self):
        """Test that dynamic class generation when the superclass init does not include all fields works"""

        class FixedAttrBar(Bar):
            @docval({'name': 'name', 'type': str, 'doc': 'the name of this Bar'},
                    {'name': 'data', 'type': ('data', 'array_data'), 'doc': 'some data'},
                    {'name': 'attr2', 'type': int, 'doc': 'another attribute'},
                    {'name': 'attr3', 'type': float, 'doc': 'a third attribute', 'default': 3.14},
                    {'name': 'foo', 'type': 'Foo', 'doc': 'a group', 'default': None})
            def __init__(self, **kwargs):
                kwargs["attr1"] = "fixed_attr1"
                super().__init__(**kwargs)

        # overwrite the "Bar" to Bar class mapping from setUp()
        self.type_map.register_container_type(CORE_NAMESPACE, "Bar", FixedAttrBar)

        baz_spec = GroupSpec('A test extension with no Container class',
                             data_type_def='Baz', data_type_inc=self.bar_spec,
                             attributes=[AttributeSpec('attr3', 'a float attribute', 'float'),
                                         AttributeSpec('attr4', 'another float attribute', 'float')])
        self.spec_catalog.register_spec(baz_spec, 'extension.yaml')
        cls = self.type_map.get_dt_container_cls('Baz', CORE_NAMESPACE)
        expected_args = {'name', 'data', 'attr2', 'attr3', 'attr4'}
        received_args = set()
        for x in get_docval(cls.__init__):
            if x['name'] != 'foo':
                received_args.add(x['name'])
                with self.subTest(name=x['name']):
                    self.assertNotIn('default', x)
        self.assertSetEqual(expected_args, received_args)
        self.assertTrue(issubclass(cls, FixedAttrBar))
        inst = cls(name="My Baz", data=[1, 2, 3, 4], attr2=1000, attr3=98.6, attr4=1.0)
        self.assertEqual(inst.attr1, "fixed_attr1")

    def test_multi_container_spec(self):
        multi_spec = GroupSpec(
            doc='A test extension that contains a multi',
            data_type_def='Multi',
            groups=[
                GroupSpec(data_type_inc=self.bar_spec, doc='test multi', quantity='*')
            ],
            attributes=[
                AttributeSpec(name='attr3', doc='a float attribute', dtype='float')
            ]
        )
        self.spec_catalog.register_spec(multi_spec, 'extension.yaml')
        Bar = self.type_map.get_dt_container_cls('Bar', CORE_NAMESPACE)
        Multi = self.type_map.get_dt_container_cls('Multi', CORE_NAMESPACE)
        assert issubclass(Multi, MultiContainerInterface)
        assert Multi.__clsconf__ == [
            dict(
                attr='bars',
                type=Bar,
                add='add_bars',
                get='get_bars',
                create='create_bars'
            )
        ]

        multi = Multi(
            name='my_multi',
            bars=[Bar(name='my_bar', data=list(range(10)), attr1='value1', attr2=10)],
            attr3=5.
        )
        assert multi.bars['my_bar'] == Bar(name='my_bar', data=list(range(10)), attr1='value1', attr2=10)
        assert multi.attr3 == 5.

    def test_multi_container_spec_with_inc(self):
        multi_spec = GroupSpec(
            doc='A test extension that contains a multi',
            data_type_def='Multi',
            data_type_inc=self.bar_spec,
            groups=[
                GroupSpec(data_type_inc=self.bar_spec, doc='test multi', quantity='*')
            ],
            attributes=[
                AttributeSpec(name='attr3', doc='a float attribute', dtype='float')
            ]
        )
        self.spec_catalog.register_spec(multi_spec, 'extension.yaml')
        Bar = self.type_map.get_dt_container_cls('Bar', CORE_NAMESPACE)
        Multi = self.type_map.get_dt_container_cls('Multi', CORE_NAMESPACE)
        assert issubclass(Multi, MultiContainerInterface)
        assert issubclass(Multi, Bar)
        assert Multi.__mro__ == (Multi, Bar, MultiContainerInterface, Container, AbstractContainer, object)
        assert Multi.__clsconf__ == [
            dict(
                attr='bars',
                type=Bar,
                add='add_bars',
                get='get_bars',
                create='create_bars'
            )
        ]

        multi = Multi(
            name='my_multi',
            bars=[Bar(name='my_bar', data=list(range(10)), attr1='value1', attr2=10)],
            data=list(range(10)),  # from data_type_inc: Bar
            attr1='value1',  # from data_type_inc: Bar
            attr2=10,  # from data_type_inc: Bar
            attr3=5.
        )
        assert multi.data == list(range(10))
        assert multi.attr1 == 'value1'
        assert multi.attr2 == 10
        assert multi.bars['my_bar'] == Bar(name='my_bar', data=list(range(10)), attr1='value1', attr2=10)
        assert multi.attr3 == 5.

    def test_multi_container_spec_optional(self):
        multi_spec = GroupSpec(
            doc='A test extension that contains a multi',
            data_type_def='Multi',
            groups=[
                GroupSpec(data_type_inc=self.bar_spec, doc='test multi', quantity='*')
            ],
            attributes=[
                AttributeSpec(name='attr3', doc='a float attribute', dtype='float')
            ]
        )
        self.spec_catalog.register_spec(multi_spec, 'extension.yaml')
        Multi = self.type_map.get_dt_container_cls('Multi', CORE_NAMESPACE)
        multi = Multi(
            name='my_multi',
            attr3=5.
        )
        assert len(multi.bars) == 0


class TestGetClassSeparateNamespace(TestCase):

    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        if os.path.exists(self.test_dir):  # start clean
            self.tearDown()
        os.mkdir(self.test_dir)

        self.bar_spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='Bar',
            datasets=[
                DatasetSpec(name='data', doc='a dataset', dtype='int')
            ],
            attributes=[
                AttributeSpec(name='attr1', doc='a string attribute', dtype='text'),
                AttributeSpec(name='attr2', doc='an integer attribute', dtype='int')
            ]
        )
        self.type_map = TypeMap()
        create_load_namespace_yaml(
            namespace_name=CORE_NAMESPACE,
            specs=[self.bar_spec],
            output_dir=self.test_dir,
            incl_types=dict(),
            type_map=self.type_map
        )

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_get_class_separate_ns(self):
        """Test that get_class correctly sets the name and type hierarchy across namespaces."""
        self.type_map.register_container_type(CORE_NAMESPACE, 'Bar', Bar)
        baz_spec = GroupSpec(
            doc='A test extension',
            data_type_def='Baz',
            data_type_inc='Bar',
        )
        create_load_namespace_yaml(
            namespace_name='ndx-test',
            specs=[baz_spec],
            output_dir=self.test_dir,
            incl_types={CORE_NAMESPACE: ['Bar']},
            type_map=self.type_map
        )

        cls = self.type_map.get_dt_container_cls('Baz', 'ndx-test')
        self.assertEqual(cls.__name__, 'Baz')
        self.assertTrue(issubclass(cls, Bar))

    def _build_separate_namespaces(self):
        # create an empty extension to test ClassGenerator._get_container_type resolution
        # the Bar class has not been mapped yet to the bar spec
        qux_spec = DatasetSpec(
            doc='A test extension',
            data_type_def='Qux'
        )
        spam_spec = DatasetSpec(
            doc='A test extension',
            data_type_def='Spam'
        )
        create_load_namespace_yaml(
            namespace_name='ndx-qux',
            specs=[qux_spec, spam_spec],
            output_dir=self.test_dir,
            incl_types={},
            type_map=self.type_map
        )
        # resolve Spam first so that ndx-qux is resolved first
        self.type_map.get_dt_container_cls('Spam', 'ndx-qux')

        baz_spec = GroupSpec(
            doc='A test extension',
            data_type_def='Baz',
            data_type_inc='Bar',
            groups=[
                GroupSpec(data_type_inc='Qux', doc='a qux', quantity='?'),
                GroupSpec(data_type_inc='Bar', doc='a bar', quantity='?')
            ]
        )
        create_load_namespace_yaml(
            namespace_name='ndx-test',
            specs=[baz_spec],
            output_dir=self.test_dir,
            incl_types={
                CORE_NAMESPACE: ['Bar'],
                'ndx-qux': ['Qux']
            },
            type_map=self.type_map
        )

    def _check_classes(self, baz_cls, bar_cls, bar_cls2, qux_cls, qux_cls2):
        self.assertEqual(qux_cls.__name__, 'Qux')
        self.assertEqual(baz_cls.__name__, 'Baz')
        self.assertEqual(bar_cls.__name__, 'Bar')
        self.assertIs(bar_cls, bar_cls2)  # same class, two different namespaces
        self.assertIs(qux_cls, qux_cls2)
        self.assertTrue(issubclass(qux_cls, Data))
        self.assertTrue(issubclass(baz_cls, bar_cls))
        self.assertTrue(issubclass(bar_cls, Container))

        qux_inst = qux_cls(name='qux_name', data=[1])
        bar_inst = bar_cls(name='bar_name', data=100, attr1='a string', attr2=10)
        baz_inst = baz_cls(name='baz_name', qux=qux_inst, bar=bar_inst, data=100, attr1='a string', attr2=10)
        self.assertIs(baz_inst.qux, qux_inst)

    def test_get_class_include_from_separate_ns_1(self):
        """Test that get_class correctly sets the name and includes types correctly across namespaces.
        This is one of multiple tests carried out to ensure that order of which get_dt_container_cls is called
        does not impact the results

        first use EXTENSION namespace, then use ORIGINAL namespace
        """
        self._build_separate_namespaces()

        baz_cls = self.type_map.get_dt_container_cls('Baz', 'ndx-test')  # Qux and Bar are not yet resolved
        bar_cls = self.type_map.get_dt_container_cls('Bar', 'ndx-test')
        bar_cls2 = self.type_map.get_dt_container_cls('Bar', CORE_NAMESPACE)
        qux_cls = self.type_map.get_dt_container_cls('Qux', 'ndx-test')
        qux_cls2 = self.type_map.get_dt_container_cls('Qux', 'ndx-qux')

        self._check_classes(baz_cls, bar_cls, bar_cls2, qux_cls, qux_cls2)

    def test_get_class_include_from_separate_ns_2(self):
        """Test that get_class correctly sets the name and includes types correctly across namespaces.
        This is one of multiple tests carried out to ensure that order of which get_dt_container_cls is called
        does not impact the results

        first use ORIGINAL namespace, then use EXTENSION namespace
        """
        self._build_separate_namespaces()

        baz_cls = self.type_map.get_dt_container_cls('Baz', 'ndx-test')  # Qux and Bar are not yet resolved
        bar_cls2 = self.type_map.get_dt_container_cls('Bar', CORE_NAMESPACE)
        bar_cls = self.type_map.get_dt_container_cls('Bar', 'ndx-test')
        qux_cls = self.type_map.get_dt_container_cls('Qux', 'ndx-test')
        qux_cls2 = self.type_map.get_dt_container_cls('Qux', 'ndx-qux')

        self._check_classes(baz_cls, bar_cls, bar_cls2, qux_cls, qux_cls2)

    def test_get_class_include_from_separate_ns_3(self):
        """Test that get_class correctly sets the name and includes types correctly across namespaces.
        This is one of multiple tests carried out to ensure that order of which get_dt_container_cls is called
        does not impact the results

        first use EXTENSION namespace, then use EXTENSION namespace
        """
        self._build_separate_namespaces()

        baz_cls = self.type_map.get_dt_container_cls('Baz', 'ndx-test')  # Qux and Bar are not yet resolved
        bar_cls = self.type_map.get_dt_container_cls('Bar', 'ndx-test')
        bar_cls2 = self.type_map.get_dt_container_cls('Bar', CORE_NAMESPACE)
        qux_cls2 = self.type_map.get_dt_container_cls('Qux', 'ndx-qux')
        qux_cls = self.type_map.get_dt_container_cls('Qux', 'ndx-test')

        self._check_classes(baz_cls, bar_cls, bar_cls2, qux_cls, qux_cls2)

    def test_get_class_include_from_separate_ns_4(self):
        """Test that get_class correctly sets the name and includes types correctly across namespaces.
        This is one of multiple tests carried out to ensure that order of which get_dt_container_cls is called
        does not impact the results

        first use ORIGINAL namespace, then use EXTENSION namespace
        """
        self._build_separate_namespaces()

        baz_cls = self.type_map.get_dt_container_cls('Baz', 'ndx-test')  # Qux and Bar are not yet resolved
        bar_cls2 = self.type_map.get_dt_container_cls('Bar', CORE_NAMESPACE)
        bar_cls = self.type_map.get_dt_container_cls('Bar', 'ndx-test')
        qux_cls2 = self.type_map.get_dt_container_cls('Qux', 'ndx-qux')
        qux_cls = self.type_map.get_dt_container_cls('Qux', 'ndx-test')

        self._check_classes(baz_cls, bar_cls, bar_cls2, qux_cls, qux_cls2)


class EmptyBar(Container):
    pass


class TestBaseProcessFieldSpec(TestCase):

    def setUp(self):
        self.bar_spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='EmptyBar'
        )
        self.spec_catalog = SpecCatalog()
        self.spec_catalog.register_spec(self.bar_spec, 'test.yaml')
        self.namespace = SpecNamespace('a test namespace', CORE_NAMESPACE,
                                       [{'source': 'test.yaml'}],
                                       version='0.1.0',
                                       catalog=self.spec_catalog)
        self.namespace_catalog = NamespaceCatalog()
        self.namespace_catalog.add_namespace(CORE_NAMESPACE, self.namespace)
        self.type_map = TypeMap(self.namespace_catalog)
        self.type_map.register_container_type(CORE_NAMESPACE, 'EmptyBar', EmptyBar)

    def test_update_docval(self):
        """Test update_docval_args for a variety of data types and mapping configurations."""
        spec = GroupSpec(
            doc="A test group specification with a data type",
            data_type_def="Baz",
            groups=[
                GroupSpec(doc="a group", data_type_inc="EmptyBar", quantity="?")
            ],
            datasets=[
                DatasetSpec(
                    doc="a dataset",
                    dtype="int",
                    name="data",
                    attributes=[
                        AttributeSpec(name="attr2", doc="an integer attribute", dtype="int")
                    ],
                )
            ],
            attributes=[
                AttributeSpec(name="attr1", doc="a string attribute", dtype="text"),
                AttributeSpec(name="attr3", doc="a numeric attribute", dtype="numeric"),
                AttributeSpec(name="attr4", doc="a float attribute", dtype="float"),
            ],
        )

        expected = [
            {'name': 'data', 'type': (int, np.int32, np.int64), 'doc': 'a dataset'},
            {'name': 'attr1', 'type': str, 'doc': 'a string attribute'},
            {'name': 'attr2', 'type': (int, np.int32, np.int64), 'doc': 'an integer attribute'},
            {'name': 'attr3', 'doc': 'a numeric attribute',
             'type': (float, np.float32, np.float64, np.int8, np.int16,
                      np.int32, np.int64, int, np.uint8, np.uint16,
                      np.uint32, np.uint64)},
            {'name': 'attr4', 'doc': 'a float attribute',
             'type': (float, np.float32, np.float64)},
            {'name': 'bar', 'type': EmptyBar, 'doc': 'a group', 'default': None},
        ]

        not_inherited_fields = {
            'data': spec.get_dataset('data'),
            'attr1': spec.get_attribute('attr1'),
            'attr2': spec.get_dataset('data').get_attribute('attr2'),
            'attr3': spec.get_attribute('attr3'),
            'attr4': spec.get_attribute('attr4'),
            'bar': spec.groups[0]
        }

        docval_args = list()
        for i, attr_name in enumerate(not_inherited_fields):
            with self.subTest(attr_name=attr_name):
                CustomClassGenerator.process_field_spec(
                    classdict={},
                    docval_args=docval_args,
                    parent_cls=EmptyBar,  # <-- arbitrary class
                    attr_name=attr_name,
                    not_inherited_fields=not_inherited_fields,
                    type_map=self.type_map,
                    spec=spec
                )
                self.assertListEqual(docval_args, expected[:(i+1)])  # compare with the first i elements of expected

    def test_update_docval_attr_shape(self):
        """Test that update_docval_args for an attribute with shape sets the type and shape keys."""
        spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='Baz',
            attributes=[
                AttributeSpec(name='attr1', doc='a string attribute', dtype='text', shape=[None])
            ]
        )
        not_inherited_fields = {'attr1': spec.get_attribute('attr1')}

        docval_args = list()
        CustomClassGenerator.process_field_spec(
            classdict={},
            docval_args=docval_args,
            parent_cls=EmptyBar,  # <-- arbitrary class
            attr_name='attr1',
            not_inherited_fields=not_inherited_fields,
            type_map=TypeMap(),
            spec=spec
        )

        expected = [{'name': 'attr1', 'type': ('array_data', 'data'), 'doc': 'a string attribute', 'shape': [None]}]
        self.assertListEqual(docval_args, expected)

    def test_update_docval_dset_shape(self):
        """Test that update_docval_args for a dataset with shape sets the type and shape keys."""
        spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='Baz',
            datasets=[
                DatasetSpec(name='dset1', doc='a string dataset', dtype='text', shape=[None])
            ]
        )
        not_inherited_fields = {'dset1': spec.get_dataset('dset1')}

        docval_args = list()
        CustomClassGenerator.process_field_spec(
            classdict={},
            docval_args=docval_args,
            parent_cls=EmptyBar,  # <-- arbitrary class
            attr_name='dset1',
            not_inherited_fields=not_inherited_fields,
            type_map=TypeMap(),
            spec=spec
        )

        expected = [{'name': 'dset1', 'type': ('array_data', 'data'), 'doc': 'a string dataset', 'shape': [None]}]
        self.assertListEqual(docval_args, expected)

    def test_update_docval_default_value(self):
        """Test that update_docval_args for an optional field with default value sets the default key."""
        spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='Baz',
            attributes=[
                AttributeSpec(name='attr1', doc='a string attribute', dtype='text', required=False,
                              default_value='value')
            ]
        )
        not_inherited_fields = {'attr1': spec.get_attribute('attr1')}

        docval_args = list()
        CustomClassGenerator.process_field_spec(
            classdict={},
            docval_args=docval_args,
            parent_cls=EmptyBar,  # <-- arbitrary class
            attr_name='attr1',
            not_inherited_fields=not_inherited_fields,
            type_map=TypeMap(),
            spec=spec
        )

        expected = [{'name': 'attr1', 'type': str, 'doc': 'a string attribute', 'default': 'value'}]
        self.assertListEqual(docval_args, expected)

    def test_update_docval_default_value_none(self):
        """Test that update_docval_args for an optional field sets default: None."""
        spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='Baz',
            attributes=[
                AttributeSpec(name='attr1', doc='a string attribute', dtype='text', required=False)
            ]
        )
        not_inherited_fields = {'attr1': spec.get_attribute('attr1')}

        docval_args = list()
        CustomClassGenerator.process_field_spec(
            classdict={},
            docval_args=docval_args,
            parent_cls=EmptyBar,  # <-- arbitrary class
            attr_name='attr1',
            not_inherited_fields=not_inherited_fields,
            type_map=TypeMap(),
            spec=spec
        )

        expected = [{'name': 'attr1', 'type': str, 'doc': 'a string attribute', 'default': None}]
        self.assertListEqual(docval_args, expected)

    def test_update_docval_default_value_none_required_parent(self):
        """Test that update_docval_args for an optional field with a required parent sets default: None."""
        spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='Baz',
            groups=[
                GroupSpec(
                    name='group1',
                    doc='required untyped group',
                    attributes=[
                        AttributeSpec(name='attr1', doc='a string attribute', dtype='text', required=False)
                    ]
                )
            ]
        )
        not_inherited_fields = {'attr1': spec.get_group('group1').get_attribute('attr1')}

        docval_args = list()
        CustomClassGenerator.process_field_spec(
            classdict={},
            docval_args=docval_args,
            parent_cls=EmptyBar,  # <-- arbitrary class
            attr_name='attr1',
            not_inherited_fields=not_inherited_fields,
            type_map=TypeMap(),
            spec=spec
        )

        expected = [{'name': 'attr1', 'type': str, 'doc': 'a string attribute', 'default': None}]
        self.assertListEqual(docval_args, expected)

    def test_update_docval_required_field_optional_parent(self):
        """Test that update_docval_args for a required field with an optional parent sets default: None."""
        spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='Baz',
            groups=[
                GroupSpec(
                    name='group1',
                    doc='required untyped group',
                    attributes=[
                        AttributeSpec(name='attr1', doc='a string attribute', dtype='text')
                    ],
                    quantity='?'
                )
            ]
        )
        not_inherited_fields = {'attr1': spec.get_group('group1').get_attribute('attr1')}

        docval_args = list()
        CustomClassGenerator.process_field_spec(
            classdict={},
            docval_args=docval_args,
            parent_cls=EmptyBar,  # <-- arbitrary class
            attr_name='attr1',
            not_inherited_fields=not_inherited_fields,
            type_map=TypeMap(),
            spec=spec
        )

        expected = [{'name': 'attr1', 'type': str, 'doc': 'a string attribute', 'default': None}]
        self.assertListEqual(docval_args, expected)

    def test_process_field_spec_overwrite(self):
        """Test that docval generation overwrites previous docval args."""
        spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='Baz',
            attributes=[
                AttributeSpec(name='attr1', doc='a string attribute', dtype='text', shape=[None])
            ]
        )
        not_inherited_fields = {'attr1': spec.get_attribute('attr1')}

        docval_args = [{'name': 'attr1', 'type': ('array_data', 'data'), 'doc': 'a string attribute',
                        'shape': [[None], [None, None]]},  # this dict will be overwritten below
                       {'name': 'attr2', 'type': ('array_data', 'data'), 'doc': 'a string attribute',
                        'shape': [[None], [None, None]]}]
        CustomClassGenerator.process_field_spec(
            classdict={},
            docval_args=docval_args,
            parent_cls=EmptyBar,  # <-- arbitrary class
            attr_name='attr1',
            not_inherited_fields=not_inherited_fields,
            type_map=TypeMap(),
            spec=spec
        )

        expected = [{'name': 'attr1', 'type': ('array_data', 'data'), 'doc': 'a string attribute',
                    'shape': [None]},
                    {'name': 'attr2', 'type': ('array_data', 'data'), 'doc': 'a string attribute',
                     'shape': [[None], [None, None]]}]
        self.assertListEqual(docval_args, expected)

    def test_process_field_spec_link(self):
        """Test that processing a link spec does not set child=True in __fields__."""
        classdict = {}
        not_inherited_fields = {'attr3': LinkSpec(name='attr3', target_type='EmptyBar', doc='a link')}
        CustomClassGenerator.process_field_spec(
            classdict=classdict,
            docval_args=[],
            parent_cls=EmptyBar,  # <-- arbitrary class
            attr_name='attr3',
            not_inherited_fields=not_inherited_fields,
            type_map=self.type_map,
            spec=GroupSpec('dummy', 'doc')
        )

        expected = {'__fields__': [{'name': 'attr3', 'doc': 'a link'}]}
        self.assertDictEqual(classdict, expected)

    def test_post_process_fixed_name(self):
        """Test that docval generation for a class with a fixed name does not contain a docval arg for name."""
        spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='Baz',
            name='MyBaz',  # <-- fixed name
            attributes=[
                AttributeSpec(
                    name='attr1',
                    doc='a string attribute',
                    dtype='text',
                    shape=[None]
                )
            ]
        )

        classdict = {}
        bases = [Container]
        docval_args = [{'name': 'name', 'type': str, 'doc': 'name'},
                       {'name': 'attr1', 'type': ('array_data', 'data'), 'doc': 'a string attribute',
                        'shape': [None]}]
        CustomClassGenerator.post_process(classdict, bases, docval_args, spec)

        expected = [{'name': 'attr1', 'type': ('array_data', 'data'), 'doc': 'a string attribute',
                     'shape': [None]}]
        self.assertListEqual(docval_args, expected)

    def test_post_process_default_name(self):
        """Test that docval generation for a class with a default name has the default value for name set."""
        spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='Baz',
            default_name='MyBaz',  # <-- default name
            attributes=[
                AttributeSpec(
                    name='attr1',
                    doc='a string attribute',
                    dtype='text',
                    shape=[None]
                )
            ]
        )

        classdict = {}
        bases = [Container]
        docval_args = [{'name': 'name', 'type': str, 'doc': 'name'},
                       {'name': 'attr1', 'type': ('array_data', 'data'), 'doc': 'a string attribute',
                        'shape': [None]}]
        CustomClassGenerator.post_process(classdict, bases, docval_args, spec)

        expected = [{'name': 'name', 'type': str, 'doc': 'name', 'default': 'MyBaz'},
                    {'name': 'attr1', 'type': ('array_data', 'data'), 'doc': 'a string attribute',
                     'shape': [None]}]
        self.assertListEqual(docval_args, expected)


class TestMCIProcessFieldSpec(TestCase):

    def setUp(self):
        bar_spec = GroupSpec(
            doc='A test group specification with a data type',
            data_type_def='EmptyBar'
        )
        specs = [bar_spec]
        container_classes = {'EmptyBar': EmptyBar}
        self.type_map = create_test_type_map(specs, container_classes)

    def test_update_docval(self):
        spec = GroupSpec(data_type_inc='EmptyBar', doc='test multi', quantity='*')
        classdict = dict()
        docval_args = []
        not_inherited_fields = {'empty_bars': spec}
        MCIClassGenerator.process_field_spec(
            classdict=classdict,
            docval_args=docval_args,
            parent_cls=Container,
            attr_name='empty_bars',
            not_inherited_fields=not_inherited_fields,
            type_map=self.type_map,
            spec=spec
        )

        expected = [
            dict(
                attr='empty_bars',
                type=EmptyBar,
                add='add_empty_bars',
                get='get_empty_bars',
                create='create_empty_bars'
            )
        ]
        self.assertEqual(classdict['__clsconf__'], expected)

    def test_update_init_zero_or_more(self):
        spec = GroupSpec(data_type_inc='EmptyBar', doc='test multi', quantity='*')
        classdict = dict()
        docval_args = []
        not_inherited_fields = {'empty_bars': spec}
        MCIClassGenerator.process_field_spec(
            classdict=classdict,
            docval_args=docval_args,
            parent_cls=Container,
            attr_name='empty_bars',
            not_inherited_fields=not_inherited_fields,
            type_map=self.type_map,
            spec=spec
        )

        expected = [{'name': 'empty_bars', 'type': (list, tuple, dict, EmptyBar), 'doc': 'test multi', 'default': None}]
        self.assertListEqual(docval_args, expected)

    def test_update_init_one_or_more(self):
        spec = GroupSpec(data_type_inc='EmptyBar', doc='test multi', quantity='+')
        classdict = dict()
        docval_args = []
        not_inherited_fields = {'empty_bars': spec}
        MCIClassGenerator.process_field_spec(
            classdict=classdict,
            docval_args=docval_args,
            parent_cls=Bar,
            attr_name='empty_bars',
            not_inherited_fields=not_inherited_fields,
            type_map=self.type_map,
            spec=spec
        )

        expected = [{'name': 'empty_bars', 'type': (list, tuple, dict, EmptyBar), 'doc': 'test multi'}]
        self.assertListEqual(docval_args, expected)

    def test_post_process(self):
        multi_spec = GroupSpec(
            doc='A test extension that contains a multi',
            data_type_def='Multi',
            groups=[
                GroupSpec(data_type_inc='EmptyBar', doc='test multi', quantity='*')
            ],
        )
        classdict = dict(
            __clsconf__=[
                dict(
                    attr='empty_bars',
                    type=EmptyBar,
                    add='add_empty_bars',
                    get='get_empty_bars',
                    create='create_empty_bars'
                )
            ]
        )
        bases = [Bar]
        docval_args = []
        MCIClassGenerator.post_process(classdict, bases, docval_args, multi_spec)
        self.assertEqual(bases, [Bar, MultiContainerInterface])

    def test_post_process_already_multi(self):
        class Multi1(MultiContainerInterface):
            pass

        multi_spec = GroupSpec(
            doc='A test extension that contains a multi and extends a multi',
            data_type_def='Multi2',
            data_type_inc='Multi1',
            groups=[
                GroupSpec(data_type_inc='EmptyBar', doc='test multi', quantity='*')
            ],
        )
        classdict = dict(
            __clsconf__=[
                dict(
                    attr='empty_bars',
                    type=EmptyBar,
                    add='add_empty_bars',
                    get='get_empty_bars',
                    create='create_empty_bars'
                )
            ]
        )
        bases = [Multi1]
        docval_args = []
        MCIClassGenerator.post_process(classdict, bases, docval_args, multi_spec)
        self.assertEqual(bases, [Multi1])

    def test_post_process_container(self):
        class Multi1(MultiContainerInterface):
            pass

        multi_spec = GroupSpec(
            doc='A test extension that contains a multi and extends a multi',
            data_type_def='Multi1',
            groups=[
                GroupSpec(data_type_inc='EmptyBar', doc='test multi', quantity='*')
            ],
        )
        classdict = dict(
            __clsconf__=[
                dict(
                    attr='empty_bars',
                    type=EmptyBar,
                    add='add_empty_bars',
                    get='get_empty_bars',
                    create='create_empty_bars'
                )
            ]
        )
        bases = [Container]
        docval_args = []
        MCIClassGenerator.post_process(classdict, bases, docval_args, multi_spec)
        self.assertEqual(bases, [MultiContainerInterface, Container])



from numpy import dtype

from ..spec.spec import DtypeHelper
from ..utils import docval, getargs

__all__ = [
    "Error",
    "DtypeError",
    "MissingError",
    "ExpectedArrayError",
    "ShapeError",
    "MissingDataType",
    "IllegalLinkError",
    "IncorrectDataType",
    "IncorrectQuantityError"
]


class Error:

    @docval({'name': 'name', 'type': str, 'doc': 'the name of the component that is erroneous'},
            {'name': 'reason', 'type': str, 'doc': 'the reason for the error'},
            {'name': 'location', 'type': str, 'doc': 'the location of the error', 'default': None})
    def __init__(self, **kwargs):
        self.__name = getargs('name', kwargs)
        self.__reason = getargs('reason', kwargs)
        self.__location = getargs('location', kwargs)

    @property
    def name(self):
        return self.__name

    @property
    def reason(self):
        return self.__reason

    @property
    def location(self):
        return self.__location

    @location.setter
    def location(self, loc):
        self.__location = loc

    def __str__(self):
        return self.__format_str(self.name, self.location, self.reason)

    @staticmethod
    def __format_str(name, location, reason):
        if location is not None:
            return "%s (%s): %s" % (name, location, reason)
        else:
            return "%s: %s" % (name, reason)

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        """Returns the hash value of this Error

        Note: if the location property is set after creation, the hash value will
        change. Therefore, it is important to finalize the value of location
        before getting the hash value.
        """
        return hash(self.__equatable_str())

    def __equatable_str(self):
        """A string representation of the error which can be used to check for equality

        For a single error, name can end up being different depending on whether it is
        generated from a base data type spec or from an inner type definition. These errors
        should still be considered equal because they are caused by the same problem.

        When a location is provided, we only consider the name of the field and drop the
        rest of the spec name. However, when a location is not available, then we need to
        use the fully-provided name.
        """
        if self.location is not None:
            equatable_name = self.name.split('/')[-1]
        else:
            equatable_name = self.name
        return self.__format_str(equatable_name, self.location, self.reason)

    def __eq__(self, other):
        return hash(self) == hash(other)


class DtypeError(Error):

    @docval({'name': 'name', 'type': str, 'doc': 'the name of the component that is erroneous'},
            {'name': 'expected', 'type': (dtype, type, str, list), 'doc': 'the expected dtype'},
            {'name': 'received', 'type': (dtype, type, str, list), 'doc': 'the received dtype'},
            {'name': 'location', 'type': str, 'doc': 'the location of the error', 'default': None})
    def __init__(self, **kwargs):
        name = getargs('name', kwargs)
        expected = getargs('expected', kwargs)
        received = getargs('received', kwargs)
        if isinstance(expected, list):
            expected = DtypeHelper.simplify_cpd_type(expected)
        reason = "incorrect type - expected '%s', got '%s'" % (expected, received)
        loc = getargs('location', kwargs)
        super().__init__(name, reason, location=loc)


class MissingError(Error):
    @docval({'name': 'name', 'type': str, 'doc': 'the name of the component that is erroneous'},
            {'name': 'location', 'type': str, 'doc': 'the location of the error', 'default': None})
    def __init__(self, **kwargs):
        name = getargs('name', kwargs)
        reason = "argument missing"
        loc = getargs('location', kwargs)
        super().__init__(name, reason, location=loc)


class MissingDataType(Error):
    @docval({'name': 'name', 'type': str, 'doc': 'the name of the component that is erroneous'},
            {'name': 'data_type', 'type': str, 'doc': 'the missing data type'},
            {'name': 'location', 'type': str, 'doc': 'the location of the error', 'default': None},
            {'name': 'missing_dt_name', 'type': str, 'doc': 'the name of the missing data type', 'default': None})
    def __init__(self, **kwargs):
        name, data_type, missing_dt_name = getargs('name', 'data_type', 'missing_dt_name', kwargs)
        self.__data_type = data_type
        if missing_dt_name is not None:
            reason = "missing data type %s (%s)" % (self.__data_type, missing_dt_name)
        else:
            reason = "missing data type %s" % self.__data_type
        loc = getargs('location', kwargs)
        super().__init__(name, reason, location=loc)

    @property
    def data_type(self):
        return self.__data_type


class IncorrectQuantityError(Error):
    """A validation error indicating that a child group/dataset/link has the incorrect quantity of matching elements"""
    @docval({'name': 'name', 'type': str, 'doc': 'the name of the component that is erroneous'},
            {'name': 'data_type', 'type': str, 'doc': 'the data type which has the incorrect quantity'},
            {'name': 'expected', 'type': (str, int), 'doc': 'the expected quantity'},
            {'name': 'received', 'type': (str, int), 'doc': 'the received quantity'},
            {'name': 'location', 'type': str, 'doc': 'the location of the error', 'default': None})
    def __init__(self, **kwargs):
        name, data_type, expected, received = getargs('name', 'data_type', 'expected', 'received', kwargs)
        reason = "expected a quantity of %s for data type %s, received %s" % (str(expected), data_type, str(received))
        loc = getargs('location', kwargs)
        super().__init__(name, reason, location=loc)


class ExpectedArrayError(Error):

    @docval({'name': 'name', 'type': str, 'doc': 'the name of the component that is erroneous'},
            {'name': 'expected', 'type': (tuple, list), 'doc': 'the expected shape'},
            {'name': 'received', 'type': str, 'doc': 'the received data'},
            {'name': 'location', 'type': str, 'doc': 'the location of the error', 'default': None})
    def __init__(self, **kwargs):
        name = getargs('name', kwargs)
        expected = getargs('expected', kwargs)
        received = getargs('received', kwargs)
        reason = "incorrect shape - expected an array of shape '%s', got non-array data '%s'" % (expected, received)
        loc = getargs('location', kwargs)
        super().__init__(name, reason, location=loc)


class ShapeError(Error):

    @docval({'name': 'name', 'type': str, 'doc': 'the name of the component that is erroneous'},
            {'name': 'expected', 'type': (tuple, list), 'doc': 'the expected shape'},
            {'name': 'received', 'type': (tuple, list), 'doc': 'the received shape'},
            {'name': 'location', 'type': str, 'doc': 'the location of the error', 'default': None})
    def __init__(self, **kwargs):
        name = getargs('name', kwargs)
        expected = getargs('expected', kwargs)
        received = getargs('received', kwargs)
        reason = "incorrect shape - expected '%s', got '%s'" % (expected, received)
        loc = getargs('location', kwargs)
        super().__init__(name, reason, location=loc)


class IllegalLinkError(Error):
    """
    A validation error for indicating that a link was used where an actual object
    (i.e. a dataset or a group) must be used
    """

    @docval({'name': 'name', 'type': str, 'doc': 'the name of the component that is erroneous'},
            {'name': 'location', 'type': str, 'doc': 'the location of the error', 'default': None})
    def __init__(self, **kwargs):
        name = getargs('name', kwargs)
        reason = "illegal use of link (linked object will not be validated)"
        loc = getargs('location', kwargs)
        super().__init__(name, reason, location=loc)


class IncorrectDataType(Error):
    """
    A validation error for indicating that the incorrect data_type (not dtype) was used.
    """

    @docval({'name': 'name', 'type': str, 'doc': 'the name of the component that is erroneous'},
            {'name': 'expected', 'type': str, 'doc': 'the expected data_type'},
            {'name': 'received', 'type': str, 'doc': 'the received data_type'},
            {'name': 'location', 'type': str, 'doc': 'the location of the error', 'default': None})
    def __init__(self, **kwargs):
        name = getargs('name', kwargs)
        expected = getargs('expected', kwargs)
        received = getargs('received', kwargs)
        reason = "incorrect data_type - expected '%s', got '%s'" % (expected, received)
        loc = getargs('location', kwargs)
        super().__init__(name, reason, location=loc)



"""
Module providing additional functionality for dealing with hierarchically nested tables, i.e.,
tables containing DynamicTableRegion references.
"""
import pandas as pd
import numpy as np
from hdmf.common.table import DynamicTable, DynamicTableRegion, VectorIndex
from hdmf.common.alignedtable import AlignedDynamicTable
from hdmf.utils import docval, getargs


@docval({'name': 'dynamic_table', 'type': DynamicTable,
         'doc': 'DynamicTable object to be converted to a hierarchical pandas.Dataframe'},
        returns="Hierarchical pandas.DataFrame with usually a pandas.MultiIndex on both the index and columns.",
        rtype='pandas.DataFrame',
        is_method=False)
def to_hierarchical_dataframe(dynamic_table):
    """
    Create a hierarchical pandas.DataFrame that represents all data from a collection of linked DynamicTables.

    **LIMITATIONS:** Currently this function only supports DynamicTables with a single DynamicTableRegion column.
    If a table has more than one DynamicTableRegion column then the function will expand only the
    first DynamicTableRegion column found for each table. Any additional DynamicTableRegion columns will remain
    nested.

    **NOTE:** Some useful functions for further processing of the generated
    DataFrame include:

    * pandas.DataFrame.reset_index to turn the data from the pandas.MultiIndex into columns
    * :py:meth:`~hdmf.common.hierarchicaltable.drop_id_columns` to remove all 'id' columns
    * :py:meth:`~hdmf.common.hierarchicaltable.flatten_column_index` to flatten the column index
    """
    # TODO: Need to deal with the case where we have more than one DynamicTableRegion column in a given table
    # Get the references column
    foreign_columns = dynamic_table.get_foreign_columns()
    # if table does not contain any DynamicTableRegion columns then we can just convert it to a dataframe
    if len(foreign_columns) == 0:
        return dynamic_table.to_dataframe()
    hcol_name = foreign_columns[0]   # We only denormalize the first foreign column for now
    hcol = dynamic_table[hcol_name]  # Either a VectorIndex pointing to a DynamicTableRegion or a DynamicTableRegion
    # Get the target DynamicTable that hcol is pointing to. If hcol is a VectorIndex then we first need
    # to get the target of it before we look up the table.
    hcol_target = hcol.table if isinstance(hcol, DynamicTableRegion) else hcol.target.table

    # Create the data variables we need to collect the data for our output dataframe and associated index
    index = []
    data = []
    columns = None
    index_names = None

    #  First we here get a list of DataFrames, one for each row of the column we need to process.
    #  If  hcol is a VectorIndex (i.e., our column is a ragged array of row indices), then simply loading
    #  the data from the VectorIndex will do the trick. If we have a regular DynamicTableRegion column,
    #  then we need to load the elements ourselves (using slice syntax to make sure we get DataFrames)
    #  one-row-at-a-time
    if isinstance(hcol, VectorIndex):
        rows = hcol.get(slice(None), index=False, df=True)
    else:
        rows = [hcol[i:(i+1)] for i in range(len(hcol))]
    # Retrieve the columns we need to iterate over from our input table. For AlignedDynamicTable we need to
    # use the get_colnames function instead of the colnames property to ensure we get all columns not just
    # the columns from the main table
    dynamic_table_colnames = (dynamic_table.get_colnames(include_category_tables=True, ignore_category_ids=False)
                              if isinstance(dynamic_table, AlignedDynamicTable)
                              else dynamic_table.colnames)

    # Case 1:  Our DynamicTableRegion column points to a DynamicTable that itself does not contain
    #          any DynamicTableRegion references (i.e., we have reached the end of our table hierarchy).
    #          If this is the case than we need to de-normalize the data and flatten the hierarchy
    if not hcol_target.has_foreign_columns():
        # Iterate over all rows, where each row is described by a DataFrame with one-or-more rows
        for row_index, row_df in enumerate(rows):
            # Since each row contains a pandas.DataFrame (with possible multiple rows), we
            # next need to iterate over all rows in that table to denormalize our data
            for row in row_df.itertuples(index=True):
                # Determine the column data for our row. Each selected row from our target table
                # becomes a row in our flattened table
                data.append(row)
                #  Determine the multi-index tuple for our row, consisting of: i) id of the row in this
                #  table, ii) all columns (except the hierarchical column we are flattening), and
                #  iii) the index (i.e., id) from our target row
                index_data = ([dynamic_table.id[row_index], ] +
                              [dynamic_table[row_index, colname]
                               for colname in dynamic_table_colnames if colname != hcol_name])
                index.append(tuple(index_data))

        # Determine the names for our index and columns of our output table
        # We need to do this even if our table was empty (i.e. even is len(rows)==0)
        # NOTE: While for a regular DynamicTable the "colnames" property will give us the full list of column names,
        #       for AlignedDynamicTable we need to use the get_colnames() function instead to make sure we include
        #       the category table columns as well.
        index_names = ([(dynamic_table.name, 'id')] +
                       [(dynamic_table.name, colname)
                        for colname in dynamic_table_colnames if colname != hcol_name])
        # Determine the name of our columns
        hcol_iter_columns = (hcol_target.get_colnames(include_category_tables=True, ignore_category_ids=False)
                             if isinstance(hcol_target, AlignedDynamicTable)
                             else hcol_target.colnames)
        columns = pd.MultiIndex.from_tuples([(hcol_target.name, 'id'), ] +
                                            [(hcol_target.name, c) for c in hcol_iter_columns],
                                            names=('source_table', 'label'))

    # Case 2:  Our DynamicTableRegion columns points to another table with a DynamicTableRegion, i.e.,
    #          we need to recursively resolve more levels of the table hierarchy
    else:
        # First we need to recursively flatten the hierarchy by calling 'to_hierarchical_dataframe()'
        # (i.e., this function) on the target of our hierarchical column
        hcol_hdf = to_hierarchical_dataframe(hcol_target)
        # Iterate over all rows, where each row is described by a DataFrame with one-or-more rows
        for row_index, row_df_level1 in enumerate(rows):
            # Since each row contains a pandas.DataFrame (with possible multiple rows), we
            # next need to iterate over all rows in that table to denormalize our data
            for row_df_level2 in row_df_level1.itertuples(index=True):
                # Since our target is itself a a DynamicTable with a DynamicTableRegion columns,
                # each target row itself may expand into multiple rows in the flattened hcol_hdf.
                # So we now need to look up the rows in hcol_hdf that correspond to the rows in
                # row_df_level2.
                # NOTE: In this look-up we assume that the ids (and hence the index) of
                #       each row in the table are in fact unique.
                for row_tuple_level3 in hcol_hdf.loc[[row_df_level2[0]]].itertuples(index=True):
                    # Determine the column data for our row.
                    data.append(row_tuple_level3[1:])
                    # Determine the multi-index tuple for our row,
                    index_data = ([dynamic_table.id[row_index], ] +
                                  [dynamic_table[row_index, colname]
                                   for colname in dynamic_table_colnames if colname != hcol_name] +
                                  list(row_tuple_level3[0]))
                    index.append(tuple(index_data))
        # Determine the names for our index and columns of our output table
        # We need to do this even if our table was empty (i.e. even is len(rows)==0)
        index_names = ([(dynamic_table.name, "id")] +
                       [(dynamic_table.name, colname)
                        for colname in dynamic_table_colnames if colname != hcol_name] +
                       hcol_hdf.index.names)
        columns = hcol_hdf.columns

    # Check if the index contains any unhashable types. If a table contains a VectorIndex column
    # (other than the DynamicTableRegion column) then "TypeError: unhashable type: 'list'" will
    # occur when converting the index to pd.MultiIndex. To avoid this error, we next check if any
    # of the columns in our index are of type list or np.ndarray
    unhashable_index_cols = []
    if len(index) > 0:
        unhashable_index_cols = [i for i, v in enumerate(index[0]) if isinstance(v, (list, np.ndarray))]

    # If we have any unhashable list or np.array objects in the index then update them to tuples.
    # Ideally we would detect this case when constructing the index, but it is easier to do this
    # here and it should not be much more expensive, but it requires iterating over all rows again
    if len(unhashable_index_cols) > 0:
        for i, v in enumerate(index):
            temp = list(v)
            for ci in unhashable_index_cols:
                temp[ci] = tuple(temp[ci])
            index[i] = tuple(temp)

    # Construct the pandas dataframe with the hierarchical multi-index
    multi_index = pd.MultiIndex.from_tuples(index, names=index_names)
    out_df = pd.DataFrame(data=data, index=multi_index, columns=columns)
    return out_df


def __get_col_name(col):
    """
    Internal helper function to get the actual name of a pandas DataFrame column from a
    column name that may consists of an arbitrary sequence of tuples. The function
    will return the last value of the innermost tuple.
    """
    curr_val = col
    while isinstance(curr_val, tuple):
        curr_val = curr_val[-1]
    return curr_val


def __flatten_column_name(col):
    """
    Internal helper function used to iteratively flatten a nested tuple

    :param col: Column name to flatten
    :type col: Tuple or String

    :returns: If col is a tuple then the result is a flat tuple otherwise col is returned as is
    """
    if isinstance(col, tuple):
        re = col
        while np.any([isinstance(v, tuple) for v in re]):
            temp = []
            for v in re:
                if isinstance(v, tuple):
                    temp += list(v)
                else:
                    temp += [v, ]
            re = temp
        return tuple(re)
    else:
        return col


@docval({'name': 'dataframe', 'type': pd.DataFrame,
         'doc': 'Pandas dataframe to update (usually generated by the to_hierarchical_dataframe function)'},
        {'name': 'inplace', 'type': 'bool', 'doc': 'Update the dataframe inplace or return a modified copy',
         'default': False},
        returns="pandas.DataFrame with the id columns removed",
        rtype='pandas.DataFrame',
        is_method=False)
def drop_id_columns(**kwargs):
    """
    Drop all columns named 'id' from the table.

    In case a column name is a tuple the function will drop any column for which
    the inner-most name is 'id'. The 'id' columns of DynamicTable is in many cases
    not necessary for analysis or display. This function allow us to easily filter
    all those columns.

    :raises TypeError: In case that dataframe parameter is not a pandas.Dataframe.
    """
    dataframe, inplace = getargs('dataframe', 'inplace', kwargs)
    col_name = 'id'
    drop_labels = []
    for col in dataframe.columns:
        if __get_col_name(col) == col_name:
            drop_labels.append(col)
    re = dataframe.drop(labels=drop_labels, axis=1, inplace=inplace)
    return dataframe if inplace else re


@docval({'name': 'dataframe', 'type': pd.DataFrame,
         'doc': 'Pandas dataframe to update (usually generated by the to_hierarchical_dataframe function)'},
        {'name': 'max_levels', 'type': (int, np.integer),
         'doc': 'Maximum number of levels to use in the resulting column Index. NOTE:  When '
                'limiting the number of levels the function simply removes levels from the '
                'beginning. As such, removing levels may result in columns with duplicate names.'
                'Value must be >0.',
         'default': None},
        {'name': 'inplace', 'type': 'bool', 'doc': 'Update the dataframe inplace or return a modified copy',
         'default': False},
        returns="pandas.DataFrame with a regular pandas.Index columns rather and a pandas.MultiIndex",
        rtype='pandas.DataFrame',
        is_method=False)
def flatten_column_index(**kwargs):
    """
    Flatten the column index of a pandas DataFrame.

    The functions changes the dataframe.columns from a pandas.MultiIndex to a normal Index,
    with each column usually being identified by a tuple of strings. This function is
    typically used in conjunction with DataFrames generated
    by :py:meth:`~hdmf.common.hierarchicaltable.to_hierarchical_dataframe`

    :raises ValueError: In case the num_levels is not >0
    :raises TypeError: In case that dataframe parameter is not a pandas.Dataframe.
    """
    dataframe, max_levels, inplace = getargs('dataframe', 'max_levels', 'inplace', kwargs)
    if max_levels is not None and max_levels <= 0:
        raise ValueError('max_levels must be greater than 0')
    # Compute the new column names
    col_names = [__flatten_column_name(col) for col in dataframe.columns.values]
    # Apply the max_levels filter. Make sure to do this only for columns that are actually tuples
    # in order not to accidentally shorten the actual string name of columns
    if max_levels is None:
        select_levels = slice(None)
    elif max_levels == 1:
        select_levels = -1
    else:  # max_levels > 1
        select_levels = slice(-max_levels, None)
    col_names = [col[select_levels] if isinstance(col, tuple) else col for col in col_names]
    re = dataframe if inplace else dataframe.copy()
    re.columns = col_names
    return re



from .catalog import SpecCatalog
from .namespace import NamespaceCatalog, SpecNamespace, SpecReader
from .spec import (AttributeSpec, DatasetSpec, DtypeHelper, DtypeSpec, GroupSpec, LinkSpec,
                   NAME_WILDCARD, RefSpec, Spec)
from .write import NamespaceBuilder, SpecWriter, export_spec
